"""Parser testing."""
import tla._ast as _ast
import tla._lre as _lr
import tla._pprint as _pp


def test_set_of_tuples():
    text = r'''
        A \X B \X C
        '''
    expr = _lr.parse_expr(text)
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '\\X', expr.operator
    assert len(expr.arguments) == 3, expr.arguments
    arg_a, arg_b, arg_c = expr.arguments
    assert_operator_name(arg_a, 'A')
    assert_operator_name(arg_b, 'B')
    assert_operator_name(arg_c, 'C')
    # print
    text = _pp.pformat_ast(expr)
    print(text)


def test_set_theory_operators():
    text = r'''
        A \cap B \cup C
        '''
    expr = _lr.parse_expr(text)
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '\\cup', expr.operator
    assert len(expr.arguments) == 2, len(expr.arguments)
    ab, arg_c = expr.arguments
    # expression `C`
    assert_operator_name(arg_c, 'C')
    # expression `A \cap B`
    assert ab.symbol.name == 'OPERATOR_APPLICATION', (
        ab.symbol)
    assert ab.operator == '\\cap', ab.operator
    assert len(ab.arguments) == 2, len(ab.arguments)
    arg_a, arg_b = ab.arguments
    assert_operator_name(arg_a, 'A')
    assert_operator_name(arg_b, 'B')
    # print
    text = _pp.pformat_ast(expr)
    print(text)


def test_keyword_unary_operators():
    text = r'''
        ---- MODULE unary_operators ----
        domain == DOMAIN f
        enabled == ENABLED action
        subsets == SUBSET S
        unchanged == UNCHANGED vars
        set_union == UNION R
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'unary_operators', module.name
    assert len(module.units) == 5, module.units
    # DOMAIN
    domain_def = module.units[0]
    assert_op_def(domain_def, 'domain')
    expr = domain_def.definiens
    assert_unary_op_app(expr, 'DOMAIN')
    arg, = expr.arguments
    assert_operator_name(arg, 'f')
    # ENABLED
    enabled_def = module.units[1]
    assert_op_def(enabled_def, 'enabled')
    expr = enabled_def.definiens
    assert_unary_op_app(expr, 'ENABLED')
    arg, = expr.arguments
    assert_operator_name(arg, 'action')
    # SUBSET
    subset_def = module.units[2]
    assert_op_def(subset_def, 'subsets')
    expr = subset_def.definiens
    assert_unary_op_app(expr, 'SUBSET')
    arg, = expr.arguments
    assert_operator_name(arg, 'S')
    # UNCHANGED
    unchanged_def = module.units[3]
    assert_op_def(unchanged_def, 'unchanged')
    expr = unchanged_def.definiens
    assert_unary_op_app(expr, 'UNCHANGED')
    arg, = expr.arguments
    assert_operator_name(arg, 'vars')
    # UNION
    set_union_def = module.units[4]
    assert_op_def(set_union_def, 'set_union')
    expr = set_union_def.definiens
    assert_unary_op_app(expr, 'UNION')
    arg, = expr.arguments
    assert_operator_name(arg, 'R')
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_unary_minus_as_argument():
    text = r'''
        op(-.)
        '''
    expr = _lr.parse_expr(text)
    assert_unary_op_app(expr, 'op')
    arg, = expr.arguments
    assert arg.symbol.name == 'OPERATOR_APPLICATION', (
        arg.symbol)
    assert arg.operator == '-.', arg.operator
    assert arg.arguments is None, arg.arguments
    # print
    text = _pp.pformat_ast(expr)
    print(text)


def test_unary_minus_as_operator():
    parser = _lr.ExprParser()
    text = r'''
        - 1
        '''
    expr = parser.parse(text)
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '-', expr.operator
    assert len(expr.arguments) == 1, len(expr.arguments)
    arg, = expr.arguments
    assert arg.symbol.name == 'INTEGRAL_NUMERAL', (
        arg.symbol)
    assert arg.value == '1', arg.value
    # print
    text = _pp.pformat_ast(expr)
    print(text)


def test_set_slice_operator():
    parser = _lr.ExprParser()
    text = r'''
        {x \in S: y \in R}
        '''
    expr = parser.parse(text)
    assert expr.symbol.name == 'SET_SLICE', expr.symbol
    x_in_s = expr.declaration
    y_in_r = expr.predicate
    # declarations
    assert_binary_op_app(x_in_s, r'\in')
    args = x_in_s.arguments
    x, s = args
    assert_operator_name(x, 'x')
    assert_operator_name(s, 'S')
    # predicate
    assert_binary_op_app(y_in_r, r'\in')
    args = y_in_r.arguments
    y, r = args
    assert_operator_name(y, 'y')
    assert_operator_name(r, 'R')


def test_set_comprehension_operator():
    parser = _lr.ExprParser()
    text = r'''
        {x: x \in R}
        '''
    expr = parser.parse(text)
    assert expr.symbol.name == 'SET_COMPREHENSION', (
        expr.symbol)
    assert_operator_name(expr.item, 'x')
    assert len(expr.declarations)
    x_in_r = expr.declarations[0]
    # declarations
    assert_binary_op_app(x_in_r, r'\in')
    args = x_in_r.arguments
    x, r = args
    assert_operator_name(x, 'x')
    assert_operator_name(r, 'R')


def test_always_eventually():
    parser = _lr.ExprParser()
    text = r'''
        []<> P
        '''
    expr = parser.parse(text)
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '[]', expr.operator
    assert len(expr.arguments) == 1, len(expr.arguments)
    arg, = expr.arguments
    assert arg.symbol.name == 'OPERATOR_APPLICATION', (
        arg.symbol)
    assert arg.operator == '<>', expr.operator
    assert len(arg.arguments) == 1, len(expr.arguments)
    p_op, = arg.arguments
    assert p_op.symbol.name == 'OPERATOR_APPLICATION', (
        p_op.symbol)
    assert p_op.operator == 'P', p_op.operator
    assert p_op.arguments is None, p_op.arguments
    # print
    text = _pp.pformat_ast(expr)
    print(text)


def test_eventually_always():
    parser = _lr.ExprParser()
    text = r'''
        <>[] P
        '''
    expr = parser.parse(text)
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '<>', expr.operator
    assert len(expr.arguments) == 1, len(expr.arguments)
    arg, = expr.arguments
    assert arg.symbol.name == 'OPERATOR_APPLICATION', (
        arg.symbol)
    assert arg.operator == '[]', expr.operator
    assert len(arg.arguments) == 1, len(expr.arguments)
    p_op, = arg.arguments
    assert p_op.symbol.name == 'OPERATOR_APPLICATION', (
        p_op.symbol)
    assert p_op.operator == 'P', p_op.operator
    assert p_op.arguments is None, p_op.arguments
    # print
    text = _pp.pformat_ast(expr)
    print(text)


def test_infix_in_def_of_leaf_proof():
    text = r'''
        ---- MODULE infix_then_step ----
        THEOREM TRUE
        PROOF
        <1>1. TRUE
            BY DEF +
        <1>2. TRUE
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol == 'module', module
    assert module.name == 'infix_then_step', module.name
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol == 'theorem', theorem
    assert theorem.name is None, theorem.name
    assert theorem.theorem_keyword == 'THEOREM', (
        theorem.theorem_keyword)
    goal = theorem.goal
    proof = theorem.proof
    # theorem goal `TRUE`
    assert_boolean_literal(goal, 'TRUE')
    # proof of theorem
    assert proof.symbol == 'proof', proof
    assert proof.proof_keyword is True, proof.proof_keyword
    assert len(proof.steps) == 2, len(proof.steps)
    step_11, step_12 = proof.steps
    # step `<1>1`
    assert step_11.symbol == 'proof_step', step_11.symbol
    assert step_11.proof_keyword is None, (
        step_11.proof_keyword)
    assert step_11.name == '<1>1.', step_11.name
    main = step_11.main
    assert_boolean_literal(main, 'TRUE')
    assert step_11.proof is not None, step_11.proof
    proof = step_11.proof
    # proof of step `<1>1`
    assert proof.symbol == 'by', proof
    assert proof.only is False, proof.only
    assert proof.facts is None, proof.facts
    assert len(proof.names) == 1, proof.names
    # identifier of operator `+`
    plus, = proof.names
    assert plus.symbol == 'operator_application', plus
    assert plus.operator == '+', plus.operator
    assert plus.arguments is None, plus.arguments
    # step `<1>2`
    assert step_12.symbol == 'proof_step', step_12.symbol
    assert step_12.proof_keyword is None, (
        step_12.proof_keyword)
    assert step_12.name == '<1>2.', step_12.name
    main = step_12.main
    assert_boolean_literal(main, 'TRUE')
    assert step_12.proof is None, step_12.proof
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_identifiers_in_leaf_proofs():
    text = r'''
        ---- MODULE leaf_proof ----
        COROLLARY ASSUME TRUE PROVE TRUE
        <1>2. M!Op
            BY ONLY <1>1 DEF M!Op
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'leaf_proof', module.name
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol.name == 'THEOREM', (
        theorem.symbol)
    assert theorem.name is None, theorem.name
    assert theorem.theorem_keyword == 'COROLLARY', (
        theorem.theorem_keyword)
    goal = theorem.goal
    proof = theorem.proof
    # theorem goal `TRUE`
    assert goal.symbol.name == 'SEQUENT', goal.symbol
    asms = goal.assume
    assert len(asms) == 1, len(asms)
    assume, = asms
    assert_boolean_literal(assume, 'TRUE')
    assert_boolean_literal(goal.prove, 'TRUE')
    # proof
    assert proof.symbol.name == 'PROOF', proof.symbol
    assert proof.proof_keyword is None, proof.proof_keyword
    assert len(proof.steps) == 1, len(proof.steps)
    step_12, = proof.steps
    # step `<1>2`
    assert step_12.symbol.name == 'PROOF_STEP', (
        step_12.symbol)
    assert step_12.proof_keyword is None, (
        step_12.proof_keyword)
    assert step_12.name == '<1>2.', step_12.name
    main = step_12.main
    proof = step_12.proof
    # `M!Op`
    assert (main.symbol.name ==
        'SUBEXPRESSION_REFERENCE'), main.symbol
    assert len(main.items) == 2, len(main.items)
    assert_operator_name(main.items[0], 'M')
    assert main.items[1] == 'Op', main.items
    # `BY <1>1 DEF M!Op`
    assert proof.symbol.name == 'BY', proof.symbol
    assert proof.only is True, proof.only
    facts = proof.facts
    names = proof.names
    assert len(facts) == 1, facts
    assert len(names) == 1, names
    # `<1>1`
    step_11, = facts
    assert (step_11.symbol.name ==
        'OPERATOR_APPLICATION'), step_11.symbol
    assert step_11.operator == '<1>1', step_11.operator
    assert step_11.arguments is None, step_11.arguments
    # `DEF M!Op`
    name, = names
    assert (name.symbol.name ==
        'SUBEXPRESSION_REFERENCE'), name.symbol
    assert len(name.items) == 2, len(name.items)
    assert_operator_name(name.items[0], 'M')
    assert name.items[1] == 'Op', name.items
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_suffices_proof_step():
    text = r'''
        ---- MODULE suffices_pf_step ----
        THEOREM TRUE
        <1>1. SUFFICES \A n \in Nat:  n < n + 1
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'suffices_pf_step', module.name
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol.name == 'THEOREM', (
        theorem.symbol)
    assert theorem.name is None, theorem.name
    assert theorem.theorem_keyword == 'THEOREM', (
        theorem.theorem_keyword)
    expr = theorem.goal
    proof = theorem.proof
    # `TRUE`
    assert_boolean_literal(expr, 'TRUE')
    # proof
    assert proof.symbol.name == 'PROOF', proof.symbol
    assert proof.proof_keyword is None, proof.proof_keyword
    assert len(proof.steps) == 1, proof.steps
    step_11, = proof.steps
    # step `<1>1`
    assert step_11.symbol.name == 'PROOF_STEP', (
        step_11.symbol)
    assert step_11.name == '<1>1.', step_11.name
    assert step_11.proof is None, step_11.proof
    assert step_11.proof_keyword is None, (
        step_11.proof_keyword)
    main = step_11.main
    # `SUFFICES \A n \in Nat:  n < n + 1`
    assert main.symbol.name == 'SUFFICES', main.symbol
    expr = main.predicate
    # `\A n \in Nat:  n < n + 1`
    assert expr.symbol.name == 'QUANTIFICATION', (
        expr.symbol)
    assert expr.quantifier == '\\A', expr.quantifier
    assert len(expr.declarations) == 1, expr.declarations
    decl, = expr.declarations
    pred = expr.predicate
    # `n \in Nat`
    assert_binary_op_app(decl, '\\in')
    n, nat = decl.arguments
    assert_operator_name(n, 'n')
    assert_operator_name(nat, 'Nat')
    # `n < n + 1`
    assert_binary_op_app(pred, '<')
    n, n1 = pred.arguments
    assert_operator_name(n, 'n')
    assert_binary_op_app(n1, '+')
    n, one = n1.arguments
    assert_operator_name(n, 'n')
    assert one.symbol.name == 'INTEGRAL_NUMERAL', one
    assert one.value == '1', one.value
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_quantifier_vertical_expression():
    text = r'''
        ---- MODULE quantifier_vexpr ----
        THEOREM \E n \in Nat:  /\ P(n)
                               /\ \A m \in S:  ~ P(m)
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'quantifier_vexpr', module.name
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol.name == 'THEOREM', theorem.symbol
    assert theorem.theorem_keyword == 'THEOREM', (
        theorem.theorem_keyword)
    assert theorem.name is None, theorem.name
    assert theorem.proof is None, theorem.proof
    qexpr = theorem.goal
    # `\E n \in Nat: ...`
    assert qexpr.symbol.name == 'QUANTIFICATION', (
        qexpr.symbol)
    assert qexpr.quantifier == '\\E', qexpr.quantifier
    assert len(qexpr.declarations) == 1, qexpr.declarations
    decl, = qexpr.declarations
    vexpr = qexpr.predicate
    # `n \in Nat`
    assert_binary_op_app(decl, '\\in')
    n, nat = decl.arguments
    assert_operator_name(n, 'n')
    assert_operator_name(nat, 'Nat')
    # vertical conjunction
    assert vexpr.symbol.name == 'VERTICAL_LIST', (
        vexpr.symbol)
    assert vexpr.operator == '/\\', vexpr.operator
    assert len(vexpr.arguments) == 2, vexpr.arguments
    item_1, item_2 = vexpr.arguments
    # `/\ P(n)`
    assert item_1.symbol.name == 'VERTICAL_ITEM', (
        item_1.symbol)
    expr = item_1.expression
    assert_unary_op_app(expr, 'P')
    n, = expr.arguments
    assert_operator_name(n, 'n')
    # `/\ \A m \in S:  ~ P(m)`
    assert item_2.symbol.name == 'VERTICAL_ITEM', (
        item_2.symbol)
    qexpr = item_2.expression
    assert qexpr.symbol.name == 'QUANTIFICATION', (
        qexpr.symbol)
    assert qexpr.quantifier == '\\A', qexpr.quantifier
    assert len(qexpr.declarations) == 1, qexpr.declarations
    decl, = qexpr.declarations
    pred = qexpr.predicate
    # `m \in S`
    assert_binary_op_app(decl, '\\in')
    m, nat = decl.arguments
    assert_operator_name(m, 'm')
    assert_operator_name(nat, 'S')
    # `~ P(m)`
    assert_unary_op_app(pred, '~')
    pexpr, = pred.arguments
    assert_unary_op_app(pexpr, 'P')
    m, = pexpr.arguments
    assert_operator_name(m, 'm')
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_vertical_expr_proof_step():
    text = r'''
        ---- MODULE vexpr_pf_step ----
        PROPOSITION TRUE
        <1>1.
              /\ <<e>> # << >>
            OBVIOUS
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'vexpr_pf_step', module.name
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol.name == 'THEOREM', theorem.symbol
    assert theorem.name is None, theorem.name
    assert theorem.goal is not None
    assert theorem.proof is not None
    assert theorem.theorem_keyword == 'PROPOSITION', (
        theorem.theorem_keyword)
    goal = theorem.goal
    proof = theorem.proof
    # goal `TRUE`
    assert goal.symbol.name == 'BOOLEAN_LITERAL', (
        goal.symbol)
    assert goal.value == 'TRUE', goal.value
    # proof
    assert proof.symbol.name == 'PROOF', proof.symbol
    assert proof.proof_keyword is None, (
        proof.proof_keyword)
    assert len(proof.steps) == 1, len(proof.steps)
    step_11, = proof.steps
    assert step_11.symbol.name == 'PROOF_STEP', (
        step_11.symbol)
    assert step_11.proof_keyword is None, (
        step_11.proof_keyword)
    assert step_11.name == '<1>1.', step_11.name
    main = step_11.main
    proof = step_11.proof
    # step number
    # `/\ <<e>> # << >>`
    assert main.symbol.name == 'VERTICAL_LIST', main.symbol
    assert main.operator == '/\\', main.operator
    assert len(main.arguments) == 1, main.arguments
    vitem, = main.arguments
    assert vitem.symbol.name == 'VERTICAL_ITEM', (
        vitem.symbol)
    assert vitem.operator == '/\\', vitem.operator
    expr = vitem.expression
    # `<<e>> # << >>`
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '#', expr.operator
    assert len(expr.arguments) == 2, expr.arguments
    e_tuple, empty_tuple = expr.arguments
    # `<<e>>`
    assert e_tuple.symbol.name == 'TUPLE', (
        e_tuple.symbol)
    assert len(e_tuple.items) == 1, e_tuple.items
    e, = e_tuple.items
    # `e`
    assert e.symbol.name == 'OPERATOR_APPLICATION', (
        e.symbol)
    assert e.operator == 'e', e.operator
    assert e.arguments is None, e.arguments
    # `<< >>`
    assert empty_tuple.symbol.name == 'TUPLE', (
        empty_tuple.symbol)
    assert len(empty_tuple.items) == 0, empty_tuple.items
    # `OBVIOUS`
    assert proof.symbol.name == 'OBVIOUS', proof.symbol
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_named_theorem():
    text = r'''
        ---- MODULE named_theorem ----
        LEMMA named ==
            TRUE
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'named_theorem', module.name
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol.name == 'THEOREM', theorem.symbol
    assert theorem.name == 'named', theorem.name
    assert theorem.proof is None, theorem.proof
    assert theorem.theorem_keyword == 'LEMMA', (
        theorem.theorem_keyword)
    goal = theorem.goal
    # `TRUE`
    assert goal.symbol.name == 'BOOLEAN_LITERAL', (
        goal.symbol)
    assert goal.value == 'TRUE', goal.value
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_quantifier_temporal():
    text = r'''
        ---- MODULE temporal_formula ----
        LOCAL L == \A r \in S:  SF_x(expr)
        ====
        '''
    module = _lr.parse(text)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'temporal_formula', module.name
    assert len(module.units) == 1, module.units
    defn, = module.units
    assert defn.symbol.name == 'OPERATOR_DEFINITION', (
        defn.symbol)
    assert defn.name == 'L', defn.name
    assert defn.arity is None, defn.arity
    assert defn.local is True, defn.local
    assert defn.function is None, defn.function
    qexpr = defn.definiens
    # `\A r \in S:  SF_x(expr)`
    assert qexpr.symbol.name == 'QUANTIFICATION', (
        qexpr.symbol)
    assert qexpr.quantifier == '\\A', qexpr.quantifier
    assert len(qexpr.declarations) == 1, qexpr.declarations
    decl, = qexpr.declarations
    pred = qexpr.predicate
    # `r \in S`
    assert decl.symbol.name == 'OPERATOR_APPLICATION', (
        decl.symbol)
    assert decl.operator == '\\in', decl.operator
    assert len(decl.arguments) == 2, len(decl.arguments)
    arg_r, arg_s = decl.arguments
    assert arg_r.symbol.name == 'OPERATOR_APPLICATION', (
        arg_r.symbol)
    assert arg_r.operator == 'r', arg_r.operator
    assert arg_r.arguments is None, arg_r.arguments
    assert arg_s.symbol.name == 'OPERATOR_APPLICATION', (
        arg_s.symbol)
    assert arg_s.operator == 'S', arg_s.operator
    assert arg_s.arguments is None, arg_s.arguments
    # `SF_x(expr)`
    assert pred.symbol.name == 'FAIRNESS', pred.symbol
    assert pred.operator == 'SF_', pred.operator
    x = pred.subscript
    expr = pred.action
    # subscript `x`
    assert x.symbol.name == 'OPERATOR_APPLICATION', (
        x.symbol)
    assert x.operator == 'x', x.operator
    assert x.arguments is None, x.arguments
    # action `expr`
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == 'expr', expr.operator
    assert expr.arguments is None, expr.arguments
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_unary_minus_operator_definition():
    text = r'''
        ---- MODULE binary_operator_def ----
        -. A == 0 - A
        ----
        ====
        '''
    module = _lr.parse(text)
    assert module is not None
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'binary_operator_def', module.name
    assert len(module.units) == 2, len(module.units)
    # definition
    uminus_def, dash_line = module.units
    assert dash_line == '----', dash_line
    assert uminus_def.symbol.name == 'OPERATOR_DEFINITION', (
        uminus_def.symbol)
    assert uminus_def.name == '-.', oplus_def.name
    assert uminus_def.local is None, oplus_def.local
    assert uminus_def.function is None, oplus_def.function
    arity = uminus_def.arity
    definiens = uminus_def.definiens
    # arity
    assert len(arity) == 1, arity
    param_a, = arity
    # parameter `A`
    assert param_a.symbol.name == 'OPERATOR_APPLICATION', (
        param_a.symbol)
    assert param_a.operator == 'A', param_a.operator
    assert param_a.arguments is None, param_a.arguments
    # definiens `0 - A`
    assert (definiens.symbol.name ==
        'OPERATOR_APPLICATION'), definiens.symbol
    assert definiens.operator == '-', definiens
    assert len(definiens.arguments) == 2, len(
        definiens.arguments)
    zero, arg_a = definiens.arguments
    # argument `0`
    assert zero.symbol.name == 'INTEGRAL_NUMERAL', (
        zero.symbol)
    assert zero.value == '0', zero.value
    # argument `A`
    assert arg_a.symbol.name == 'OPERATOR_APPLICATION', (
        arg_a.symbol)
    assert arg_a.operator == 'A', arg_a.operator
    assert arg_a.arguments is None, arg_a.arguments
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_binary_operator_definition():
    text = r'''
        ---- MODULE binary_operator_def ----
        LOCAL A (+) B == TRUE
        ----
        ====
        '''
    module = _lr.parse(text)
    assert module is not None
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'binary_operator_def', module.name
    assert len(module.units) == 2, len(module.units)
    # definition
    oplus_def, dash_line = module.units
    assert dash_line == '----', dash_line
    assert oplus_def.symbol.name == 'OPERATOR_DEFINITION', (
        oplus_def.symbol)
    assert oplus_def.name == '(+)', oplus_def.name
    assert oplus_def.local is True, oplus_def.local
    assert oplus_def.function is None, oplus_def.function
    arity = oplus_def.arity
    definiens = oplus_def.definiens
    # arity
    assert len(arity) == 2, arity
    param_a, param_b = arity
    # parameter `A`
    assert param_a.symbol.name == 'OPERATOR_APPLICATION', (
        param_a.symbol)
    assert param_a.operator == 'A', param_a.operator
    assert param_a.arguments is None, param_a.arguments
    # parameter `B`
    assert param_b.symbol.name == 'OPERATOR_APPLICATION', (
        param_b.symbol)
    assert param_b.operator == 'B', param_b.operator
    assert param_b.arguments is None, param_b.arguments
    # definiens `TRUE`
    assert definiens.symbol.name == 'BOOLEAN_LITERAL', (
        definiens.symbol)
    assert definiens.value == 'TRUE', definiens
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_pipe_instantiation():
    text = r'''
        ---- MODULE pipe_inst ----
        INSTANCE M WITH | <- |, x <- | x
        ====
        '''
    module = _lr.parse(text)
    assert module is not None
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'pipe_inst', module.name
    assert len(module.units) == 1, len(module.units)
    # instance
    instance, = module.units
    assert instance.symbol.name == 'INSTANCE', (
        instance.symbol)
    assert instance.name == 'M', instance.name
    assert instance.local is None, instance.local
    with_ = instance.with_substitution
    assert len(with_) == 2, len(with_)
    pipe_sub, x_sub = with_
    # `| <- |`
    assert pipe_sub.symbol.name == 'WITH_INSTANTIATION', (
        pipe_sub.symbol)
    assert pipe_sub.name == '|', pipe_sub.name
    assert pipe_sub.expression == '|', pipe_sub.expression
    # `x <- x`
    assert x_sub.symbol.name == 'WITH_INSTANTIATION', (
        x_sub.symbol)
    assert x_sub.name == 'x', x_sub.name
    vexpr = x_sub.expression
    assert vexpr.symbol.name == 'VERTICAL_LIST', (
        vexpr.symbol)
    assert vexpr.operator == '|', vexpr.operator
    assert len(vexpr.arguments) == 1, vexpr.arguments
    vitem, = vexpr.arguments
    assert vitem.symbol.name == 'VERTICAL_ITEM', (
        vitem.symbol)
    assert vitem.operator == '|', vitem.operator
    expr = vitem.expression
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == 'x', expr.operator
    assert expr.arguments is None, expr.arguments
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_nested_proofs():
    module_name = 'case_proof_step'
    source = r'''
        ---- MODULE case_proof_step ----
        THEOREM TRUE
        PROOF
        <1>1. CASE x
            PROOF
            <2>1. x = 1
            <2> QED
                BY <2>1
        <1>2. CASE ~ x
            OBVIOUS
        <1> QED
        ====
        '''
    module = _lr.parse(source)
    assert module is not None
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'case_proof_step', module.name
    assert len(module.units) == 1, len(module.units)
    # theorem
    theorem, = module.units
    assert theorem.name is None, theorem.name
    assert theorem.theorem_keyword == 'THEOREM', (
        theorem.theorem_keyword)
    goal = theorem.goal
    proof = theorem.proof
    # theorem goal TRUE
    assert goal.symbol.name == 'BOOLEAN_LITERAL', (
        goal.symbol)
    assert goal.value == 'TRUE', goal.value
    # proof
    assert proof.symbol.name == 'PROOF'
    assert proof.proof_keyword is True, proof.proof_keyword
    assert len(proof.steps) == 3, len(proof.steps)
    step_11, step_12, step_1qed = proof.steps
    # step <1>1
    assert step_11.symbol.name == 'PROOF_STEP', (
        step_11.symbol.name)
    assert step_11.proof_keyword is True, (
        step_11.proof_keyword)
    assert step_11.name == '<1>1.', step_11.name
    main = step_11.main
    proof = step_11.proof
    # goal `CASE x`
    assert main.symbol.name == 'PROOF_CASE', main.symbol
    expr = main.expression
    # expression `x`
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == 'x', expr.operator
    assert expr.arguments is None, expr.arguments
    # nested proof of <1>1
    assert proof.symbol.name == 'PROOF', proof.symbol
    assert proof.proof_keyword is None, proof.proof_keyword
    assert len(proof.steps) == 2, len(proof.steps)
    step_21, step_2qed = proof.steps
    # step <2>1
    assert step_21.symbol.name == 'PROOF_STEP', (
        step_21.symbol)
    assert step_21.proof is None, step_21.proof
    assert step_21.proof_keyword is None, (
        step_21.proof_keyword)
    assert step_21.name == '<2>1.', step_21.name
    main = step_21.main
    # goal `x = 1`
    assert main.symbol.name == 'OPERATOR_APPLICATION', (
        main.symbol)
    assert main.operator == '=', main.operator
    assert len(main.arguments) == 2, len(main.arguments)
    x, one = main.arguments
    assert x.symbol.name == 'OPERATOR_APPLICATION', (
        x.symbol)
    assert x.operator == 'x', x.operator
    assert x.arguments is None, x.arguments
    assert one.symbol.name == 'INTEGRAL_NUMERAL', (
        one.symbol)
    assert one.value == '1', one.value
    # step <2> QED
    assert step_2qed.symbol.name == 'PROOF_STEP', (
        step_2qed.symbol)
    assert step_2qed.name == '<2>', step_2qed.name
    assert step_2qed.main == 'QED', step_2qed.main
    assert step_2qed.proof is not None, step_2qed.proof
    assert step_2qed.proof_keyword is None, (
        step_1qed.proof_keyword)
    # proof `BY <2>1`
    proof = step_2qed.proof
    assert proof.symbol.name == 'BY', proof.symbol
    assert proof.only is False, proof.only
    assert proof.names is None, proof.names
    assert len(proof.facts) == 1, proof.facts
    step_name, = proof.facts
    assert (step_name.symbol.name ==
        'OPERATOR_APPLICATION'), step_name.symbol
    assert step_name.operator == '<2>1', step_name.operator
    assert step_name.arguments is None, step_name.arguments
    # step <1>2
    assert step_12.symbol.name == 'PROOF_STEP', (
        step_12.symbol.name)
    assert step_12.name == '<1>2.', step_12.name
    assert step_12.proof_keyword is None, (
        step_12.proof_keyword)
    main = step_12.main
    proof = step_12.proof
    # goal `CASE ~ x`
    assert main.symbol.name == 'PROOF_CASE', main.symbol
    expr = main.expression
    # expression `~ x`
    assert expr.symbol.name == 'OPERATOR_APPLICATION', (
        expr.symbol)
    assert expr.operator == '~', expr.operator
    assert len(expr.arguments) == 1, len(expr.arguments)
    arg, = expr.arguments
    assert arg.symbol.name == 'OPERATOR_APPLICATION', (
        arg.symbol)
    assert arg.operator == 'x', arg.operator
    assert arg.arguments is None, arg.arguments
    # proof OBVIOUS
    assert proof.symbol.name == 'OBVIOUS', proof.symbol
    # step <1> QED
    assert step_1qed.symbol.name == 'PROOF_STEP', (
        step_1qed.symbol)
    assert step_1qed.name == '<1>', step_1qed
    assert step_1qed.main == 'QED', step_1qed.main
    assert step_1qed.proof is None, step_1qed.proof
    assert step_1qed.proof_keyword is None, (
        step_1qed.proof_keyword)
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_function_definition():
    module_name = 'function_definition'
    source = r'''
        ---- MODULE function_definition ----
        f[n \in Nat] == n + 1
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 1, module
    f_def, = module.units
    # `f` definition
    assert (f_def.symbol.name ==
        'OPERATOR_DEFINITION'), f_def
    assert f_def.name == 'f', f_def
    assert f_def.arity is None, f_def
    assert f_def.local is None, f_def
    assert f_def.function is True, f_def
    # `f` definiens
    f_definiens = f_def.definiens
    assert (f_definiens.symbol.name ==
        'FUNCTION'), f_definiens
    # function constructor
    assert isinstance(
        f_definiens.declaration, list), f_definiens
    decl, = f_definiens.declaration
    expr = f_definiens.value
    # function declaration
    assert (decl.symbol.name ==
        'OPERATOR_APPLICATION'), decl
    assert decl.operator == '\\in', decl
    assert len(decl.arguments) == 2, decl
    n, nat = decl.arguments
    assert (n.symbol.name ==
        'OPERATOR_APPLICATION'), n
    assert n.arguments is None, n
    assert n.operator == 'n', n
    assert (nat.symbol.name ==
        'OPERATOR_APPLICATION'), nat
    assert nat.arguments is None, nat
    assert nat.operator == 'Nat', nat
    # function value
    assert (expr.symbol.name ==
        'OPERATOR_APPLICATION'), expr
    assert expr.operator == '+', expr
    assert len(expr.arguments) == 2, expr
    n, one = expr.arguments
    assert (n.symbol.name ==
        'OPERATOR_APPLICATION'), n
    assert n.arguments is None, n
    assert n.operator == 'n', n
    assert (one.symbol.name ==
        'INTEGRAL_NUMERAL'), one
    assert one.value == '1', one


def test_vertical_operator_end_by_infix_dedent():
    source = r'''
        ---- MODULE name ----
        A == \/ a
             \/ b
            \/ c
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == 'name', module.name
    assert len(module.units) == 1, len(module.units)
    a_def, = module.units
    # `A` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def.name
    assert a_def.arity is None, a_def.arity
    assert a_def.local is None, a_def.local
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'OPERATOR_APPLICATION'), a_definiens
    assert a_definiens.operator == r'\/', a_definiens
    assert len(a_definiens.arguments) == 2, a_definiens
    vexpr, disj = a_definiens.arguments
    # vertical expression
    assert (vexpr.symbol.name ==
        'VERTICAL_LIST'), vexpr
    assert vexpr.operator == r'\/', vexpr
    assert len(vexpr.arguments) == 2, vexpr
    a_vitem, b_vitem = vexpr.arguments
    # `\/ a`
    assert (a_vitem.symbol.name ==
        'VERTICAL_ITEM'), a_vitem
    assert a_vitem.operator == r'\/', a_vitem
    a_expr = a_vitem.expression
    assert_operator_name(a_expr, 'a')
    # `\/ b`
    assert (b_vitem.symbol.name ==
        'VERTICAL_ITEM'), b_vitem
    assert b_vitem.operator == r'\/', b_vitem
    b_expr = b_vitem.expression
    assert_operator_name(b_expr, 'b')
    # disjunct `c`
    assert_operator_name(disj, 'c')


def test_vertical_operator_contains_infix():
    source = r'''
        ---- MODULE name ----
        A == \/ a
             \/ b
              \/ c
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == 'name', module.name
    assert len(module.units) == 1, len(module.units)
    a_def, = module.units
    # `A` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def.name
    assert a_def.arity is None, a_def.arity
    assert a_def.local is None, a_def.local
    # `A` definiens
    vexpr = a_def.definiens
    # vertical expression
    assert (vexpr.symbol.name ==
        'VERTICAL_LIST'), vexpr
    assert vexpr.operator == r'\/', vexpr
    assert len(vexpr.arguments) == 2, vexpr
    a_vitem, b_vitem = vexpr.arguments
    # `\/ a`
    assert (a_vitem.symbol.name ==
        'VERTICAL_ITEM'), a_vitem
    assert a_vitem.operator == r'\/', a_vitem
    a_expr = a_vitem.expression
    assert (a_expr.symbol.name ==
        'OPERATOR_APPLICATION'), a_expr
    assert a_expr.operator == 'a', a_expr
    assert a_expr.arguments is None, a_expr
    # `\/ b \/ c`
    assert (b_vitem.symbol.name ==
        'VERTICAL_ITEM'), b_vitem
    assert b_vitem.operator == r'\/', b_vitem
    disj = b_vitem.expression
    assert (disj.symbol.name ==
        'OPERATOR_APPLICATION'), disj
    assert disj.operator == r'\/', disj
    assert len(disj.arguments) == 2, disj
    b_expr, c_expr = disj.arguments
    assert_operator_name(b_expr, 'b')
    assert_operator_name(c_expr, 'c')


def test_vertical_operator_end_by_infix_aligned():
    source = r'''
        ---- MODULE name ----
        A == \/ a
             \/ b
             /\ c
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == 'name', module.name
    assert len(module.units) == 1, len(module.units)
    a_def, = module.units
    # `A` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def.name
    assert a_def.arity is None, a_def.arity
    assert a_def.local is None, a_def.local
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'OPERATOR_APPLICATION'), a_definiens
    assert a_definiens.operator == '/\\', a_definiens
    assert len(a_definiens.arguments) == 2, a_definiens
    vexpr, disj = a_definiens.arguments
    # vertical expression
    assert (vexpr.symbol.name ==
        'VERTICAL_LIST'), vexpr
    assert vexpr.operator == r'\/', vexpr
    assert len(vexpr.arguments) == 2, vexpr
    a_vitem, b_vitem = vexpr.arguments
    # `\/ a`
    assert (a_vitem.symbol.name ==
        'VERTICAL_ITEM'), a_vitem
    assert a_vitem.operator == r'\/', a_vitem
    a_expr = a_vitem.expression
    assert_operator_name(a_expr, 'a')
    # `\/ b`
    assert (b_vitem.symbol.name ==
        'VERTICAL_ITEM'), b_vitem
    assert b_vitem.operator == r'\/', b_vitem
    b_expr = b_vitem.expression
    assert_operator_name(b_expr, 'b')
    # disjunct `c`
    assert_operator_name(disj, 'c')


def test_vertical_operators():
    module_name = 'vertical_after_infix'
    source = r'''
        ---- MODULE vertical_after_infix ----
        A == \/ b / /\ c
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 1, module
    a_def, = module.units
    # `A` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'VERTICAL_LIST'), a_definiens
    assert a_definiens.operator == r'\/', a_definiens
    assert len(a_definiens.arguments) == 1, a_definiens
    # disjunct `\/ b / /\ c`
    disjunct, = a_definiens.arguments
    assert (disjunct.symbol.name ==
        'VERTICAL_ITEM'), item
    assert disjunct.operator == r'\/', disjunct
    # expression `b / /\ c`
    expr = disjunct.expression
    assert (expr.symbol.name ==
        'OPERATOR_APPLICATION'), expr
    assert expr.operator == '/', expr
    assert len(expr.arguments) == 2, expr
    b_expr, conj = expr.arguments
    # expression `b`
    assert (b_expr.symbol.name ==
        'OPERATOR_APPLICATION'), b_expr
    assert b_expr.operator == 'b', b_expr
    assert b_expr.arguments is None, b_expr
    # expression `/\ c`
    assert (conj.symbol.name ==
        'VERTICAL_LIST'), conj
    assert conj.operator == '/\\', conj
    assert len(conj.arguments) == 1, conj
    # conjunct `/\ c`
    conjunct, = conj.arguments
    assert (conjunct.symbol.name ==
        'VERTICAL_ITEM'), conjunct
    # expression `c`
    c_expr = conjunct.expression
    assert (c_expr.symbol.name ==
        'OPERATOR_APPLICATION'), c_expr
    assert c_expr.operator == 'c', c_expr
    assert c_expr.arguments is None, c_expr


def test_preamble_comments():
    module_name = 'preamble_comments'
    module_text = r'''
        A == 1
        ---- MODULE preamble_comments ----
        LOCAL B == 1
        ====
        A == 1
        '''
    module = _lr.parse(module_text)
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 1, module
    b_def, = module.units
    # `B` definition
    assert (b_def.symbol.name ==
        'OPERATOR_DEFINITION'), b_def
    assert b_def.name == 'B', b_def
    assert b_def.arity is None, b_def
    assert b_def.local is True, b_def
    # `B` definiens
    b_definiens = b_def.definiens
    assert (b_definiens.symbol.name ==
        'INTEGRAL_NUMERAL'), b_definiens
    assert b_definiens.value == '1', b_definiens


def test_lambda_inside_undelimited():
    module_name = 'lambda_inside_undelimited'
    source = rf'''
        ---- MODULE {module_name} ----
        B(r(_)) == r(1)
        A == \A x:  B(LAMBDA y: TRUE) /\ FALSE
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 2, module
    b_def, a_def = module.units
    # `B` definition
    assert (b_def.symbol.name ==
        'OPERATOR_DEFINITION'), b_def
    assert b_def.name == 'B', b_def
    assert b_def.arity is not None, b_def
    assert b_def.local is None, b_def
    # `B` signature
    b_parameters = b_def.arity
    assert len(b_parameters) == 1, b_parameters
    # parameter `r`
    r_param, = b_parameters
    assert (r_param.symbol.name ==
        'OPERATOR_APPLICATION'), r_param
    assert r_param.operator == 'r', r_param
    assert r_param.arguments == ['_'], r_param
    # `B` definiens
    b_definiens = b_def.definiens
    assert (b_definiens.symbol.name ==
        'OPERATOR_APPLICATION'), b_definiens
    assert b_definiens.operator == 'r', b_definiens
    assert len(
        b_definiens.arguments) == 1, b_definiens
    # argument of `r(1)`
    one, = b_definiens.arguments
    assert one.symbol.name == 'INTEGRAL_NUMERAL', one
    assert one.value == '1', one
    # `B` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'QUANTIFICATION'), a_definiens
    assert a_definiens.quantifier == '\\A', a_definiens
    # quantifier declaree `x`
    a_declarations = a_definiens.declarations
    assert len(a_declarations) == 1, a_declarations
    x_declaration, = a_declarations
    assert (x_declaration.symbol.name ==
        'OPERATOR_APPLICATION'), x_declaration
    assert x_declaration.operator == 'x', x_declaration
    assert x_declaration.arguments is None, x_declaration
    # quantifier predicate
    a_predicate = a_definiens.predicate
    assert (a_predicate.symbol.name ==
        'OPERATOR_APPLICATION'), a_predicate
    assert a_predicate.operator == '/\\', a_predicate
    assert len(a_predicate.arguments) == 2, a_predicate
    arg1, arg2 = a_predicate.arguments
    # expression `B(LAMBDA y: TRUE)`
    assert (arg1.symbol.name ==
        'OPERATOR_APPLICATION'), arg1
    assert arg1.operator == 'B', arg1
    assert len(arg1.arguments) == 1, arg1
    # operator `LAMBDA y: TRUE`
    lambda_y, = arg1.arguments
    assert lambda_y.symbol.name == 'LAMBDA', lambda_y
    assert len(lambda_y.parameters) == 1, lambda_y
    # `LAMBDA` parameter: `y`
    y_param, = lambda_y.parameters
    assert (y_param.symbol.name ==
        'OPERATOR_APPLICATION'), y_param
    assert y_param.operator == 'y', y_param
    assert y_param.arguments is None, y_param
    # `LAMBDA` expression: `TRUE`
    lambda_expr = lambda_y.expression
    assert (lambda_expr.symbol.name ==
        'BOOLEAN_LITERAL'), lambda_expr
    assert lambda_expr.value == 'TRUE', lambda_expr
    # conjunct `FALSE`
    assert (arg2.symbol.name ==
        'BOOLEAN_LITERAL'), arg2
    assert arg2.value == 'FALSE', arg2


def test_nested_undelimited():
    source = r'''
        ---- MODULE let_quantification ----
        EXTENDS
            Integers

        A == LET
                b == FALSE
             IN
                \A x, y \in S:  b
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == 'let_quantification', module.name
    assert len(module.extendees) == 1, module.extendees
    integers, = module.extendees
    assert_operator_name(integers, 'Integers')
    assert len(module.units) == 1, module.units
    a_def, = module.units
    assert a_def.symbol.name == 'OPERATOR_DEFINITION', (
        a_def.symbol)
    assert a_def.name == 'A', a_def.name
    assert a_def.arity is None, a_def.arity
    assert a_def.local is None, a_def.local
    assert a_def.function is None, a_def.function
    let = a_def.definiens
    # `LET`
    assert let.symbol.name == 'LET', let.symbol
    assert len(let.definitions) == 1, let.definitions
    b_def, = let.definitions
    qexpr = let.expression
    # `b == FALSE`
    assert b_def.symbol.name == 'OPERATOR_DEFINITION', (
        b_def.symbol)
    assert b_def.arity is None, b_def.arity
    assert b_def.local is None, b_def.local
    assert b_def.function is None, b_def.function
    expr = b_def.definiens
    assert_boolean_literal(expr, 'FALSE')
    # `\A x, y \in S:  b`
    assert qexpr.symbol.name == 'QUANTIFICATION', (
        qexpr.symbol)
    assert qexpr.quantifier == '\\A', qexpr.quantifier
    assert len(qexpr.declarations) == 2, qexpr.declarations
    decls = qexpr.declarations
    pred = qexpr.predicate
    # `x, y \in S`
    x, y_in_s = decls
    assert_operator_name(x, 'x')
    assert_binary_op_app(y_in_s, '\\in')
    y, s = y_in_s.arguments
    assert_operator_name(y, 'y')
    assert_operator_name(s, 'S')
    # `b`
    assert_operator_name(pred, 'b')
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_let_expression():
    source = '''
        ---- MODULE let_expr ----
        THEOREM
            LET a == FALSE IN TRUE
        BY TRUE
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'let_expr', module.name
    assert module.extendees is None, module.extendees
    assert len(module.units) == 1, module.units
    theorem, = module.units
    assert theorem.symbol.name == 'THEOREM', theorem.symbol
    assert theorem.name is None, theorem.name
    assert theorem.theorem_keyword == 'THEOREM', (
        theorem.theorem_keyword)
    let = theorem.goal
    proof = theorem.proof
    # theorem assertion
    assert let.symbol.name == 'LET', let.symbol
    assert len(let.definitions) == 1, len(let.definitions)
    defn, = let.definitions
    in_expr = let.expression
    # `a == FALSE`
    assert defn.symbol.name == 'OPERATOR_DEFINITION', (
        defn.symbol)
    assert defn.name == 'a', defn.name
    assert defn.arity is None, defn.arity
    assert defn.local is None, defn.local
    assert defn.function is None, defn.function
    expr = defn.definiens
    assert_boolean_literal(expr, 'FALSE')
    # `TRUE`
    assert_boolean_literal(in_expr, 'TRUE')
    # proof
    assert proof.symbol.name == 'PROOF', proof.symbol
    assert proof.proof_keyword is None, proof.proof_keyword
    assert len(proof.steps) == 1, proof.steps
    leaf_proof, = proof.steps
    assert leaf_proof.symbol.name == 'BY', (
        leaf_proof.symbol)
    assert leaf_proof.only is False, leaf_proof.only
    assert leaf_proof.names is None, leaf_proof.names
    assert len(leaf_proof.facts) == 1, leaf_proof.facts
    expr, = leaf_proof.facts
    assert_boolean_literal(expr, 'TRUE')
    # print
    text = _pp.pformat_ast(module)
    print(text)


def test_definition_inside_vertical_operator():
    parser = _lr.Parser()
    # module-scope definition indented less
    # than the preceeding vertical operator
    source = r'''
        ---- MODULE smaller_indent ----
        A == /\ 1
             /\ 2

        B == 3
        ====
        '''
    module = parser.parse(source)
    _check_def_inside_vop(
        module, 'smaller_indent')
    # module-scope definition indented more
    # than the preceeding vertical operator
    source = r'''
        ---- MODULE larger_indent ----
        A == /\ 1
             /\ 2

                B == 3
        ====
        '''
    module = parser.parse(source)
    _check_def_inside_vop(
        module, 'larger_indent')
    # definition on the same line
    # with a vertical operator
    source = r'''
        ---- MODULE same_line ----
        A == /\ 1
             /\ 2     B == 3
        ====
        '''
    module = parser.parse(source)
    _check_def_inside_vop(
        module, 'same_line')
    # definition inside `LET` that
    # is within a vertical operator
    source = r'''
        ---- MODULE let_inside_vertical ----
        A == /\ LET B == 3 IN B
        ====
        '''
    module = parser.parse(source)
    _check_def_inside_let(
        module, 'let_inside_vertical')


def _check_def_inside_let(
        module, module_name):
    # module
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 1, module
    # `A` definition
    a_def, = module.units
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'VERTICAL_LIST'), a_definiens
    assert a_definiens.operator == '/\\', a_definiens
    assert len(a_definiens.arguments) == 1, a_definiens
    # first conjunct
    item, = a_definiens.arguments
    assert item.symbol.name == 'VERTICAL_ITEM', item
    assert item.operator == '/\\', item
    # `LET` expression
    let = item.expression
    assert let.symbol.name == 'LET', let
    assert len(let.definitions) == 1, let
    # `B` definition
    b_def, = let.definitions
    assert (b_def.symbol.name ==
        'OPERATOR_DEFINITION'), b_def
    assert b_def.name == 'B', b_def
    assert b_def.arity is None, b_def
    assert b_def.local is None, b_def
    # `B` definiens
    b_definiens = b_def.definiens
    assert (b_definiens.symbol.name ==
        'INTEGRAL_NUMERAL'), b_definiens
    assert b_definiens.value == '3', b_definiens
    # `IN` expression
    in_expr = let.expression
    assert (in_expr.symbol.name ==
        'OPERATOR_APPLICATION'), in_expr
    assert in_expr.operator == 'B', in_expr
    assert in_expr.arguments == None, in_expr


def _check_def_inside_vop(
        module, module_name):
    # module
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 2, module
    a_def, b_def = module.units
    # `A` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'VERTICAL_LIST'), a_definiens
    assert a_definiens.operator == '/\\', a_definiens
    assert len(a_definiens.arguments) == 2, a_definiens
    item_1, item_2 = a_definiens.arguments
    # first conjunct
    assert (item_1.symbol.name ==
        'VERTICAL_ITEM'), item_1
    assert item_1.operator == '/\\', item_1
    one = item_1.expression
    assert (one.symbol.name ==
        'INTEGRAL_NUMERAL'), one
    assert one.value == '1', one
    # second conjunct
    assert (item_2.symbol.name ==
        'VERTICAL_ITEM'), item_2
    assert item_2.operator == '/\\', item_2
    two = item_2.expression
    assert (two.symbol.name ==
        'INTEGRAL_NUMERAL'), two
    assert two.value == '2', two
    # `B` definition
    assert (b_def.symbol.name ==
        'OPERATOR_DEFINITION'), b_def
    assert b_def.name == 'B', b_def
    assert b_def.arity is None, b_def
    assert b_def.local is None, b_def
    # `B` definiens
    b_definiens = b_def.definiens
    assert (b_definiens.symbol.name ==
        'INTEGRAL_NUMERAL'), b_definiens
    assert b_definiens.value == '3', b_definiens


def test_optional_proof_keyword_define():
    # module-scope definition indented less
    # than the preceeding vertical operator
    source = '''
        ---- MODULE optional_DEFINE ----
        THEOREM TRUE
        <1> A == 1
            B == 2
        <1> QED

        C == 3
        ====
        '''
    module = _lr.parse(source)
    # module
    assert module.symbol.name == 'MODULE', module
    assert module.name == 'optional_DEFINE', module
    assert len(module.units) == 2, module
    thm, c_def = module.units
    # theorem
    assert thm.name is None, thm
    # theorem assertion
    goal = thm.goal
    assert (goal.symbol.name ==
        'BOOLEAN_LITERAL'), goal
    assert goal.value == 'TRUE', goal
    # proof
    proof = thm.proof
    assert proof.symbol.name == 'PROOF', proof
    assert len(proof.steps) == 2, proof
    # steps
    defs_step, qed_step = proof.steps
    # proof first step
    assert (defs_step.symbol.name ==
        'PROOF_STEP'), defs_step
    assert defs_step.proof is None, defs_step
    # step number of first step
    step_num = defs_step.name
    # definitions
    define_main = defs_step.main
    assert (define_main.symbol.name ==
        'DEFINE'), define_main
    assert len(
        define_main.definitions) == 2, define_main
    a_def, b_def = define_main.definitions
    # `A` definition
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'INTEGRAL_NUMERAL'), a_definiens
    assert a_definiens.value == '1', a_definiens
    # `B` definition
    assert (b_def.symbol.name ==
        'OPERATOR_DEFINITION'), b_def
    assert b_def.name == 'B', b_def
    assert b_def.arity is None, b_def
    assert b_def.local is None, b_def
    # `B` definiens
    b_definiens = b_def.definiens
    assert (b_definiens.symbol.name ==
        'INTEGRAL_NUMERAL'), b_definiens
    assert b_definiens.value == '2', b_definiens
    # QED step
    assert qed_step.main == 'QED', qed_step


def test_dot_function_application_prime():
    # The expression in this test is from
    # page 284 of the book "Specifying Systems".
    module_name = 'dot_func_prime'
    source = f'''
        ---- MODULE {module_name} ----
        A == a + b.c[d]'
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 1, module
    # `A` definition
    a_def, = module.units
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens: `a + b.c[d]'`
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'OPERATOR_APPLICATION'), a_definiens
    assert a_definiens.operator == '+', a_definiens
    assert len(a_definiens.arguments) == 2, a_definiens
    first, second = a_definiens.arguments
    # first summand: `a`
    assert (first.symbol.name ==
        'OPERATOR_APPLICATION'), first
    assert first.operator == 'a', first
    assert first.arguments is None, first
    # second summand: `b.c[d]'`
    assert (second.symbol.name ==
        'OPERATOR_APPLICATION'), second
    assert second.operator == "'", second
    assert len(second.arguments) == 1, second
    app, = second.arguments
    # function application `b.c[d]`
    assert (app.symbol.name ==
        'FUNCTION_APPLICATION'), app
    assert len(app.arguments) == 1, app
    func = app.function
    d_expr, = app.arguments
    # `d` expression
    assert (d_expr.symbol.name ==
        'OPERATOR_APPLICATION'), d_expr
    assert d_expr.operator == 'd', d_expr
    assert d_expr.arguments is None, d_expr
    # `b.c` expression
    assert func.symbol.name == 'FIELD', func
    b_expr = func.expression
    c_expr = func.name
    # expression `b`
    assert (b_expr.symbol.name ==
        'OPERATOR_APPLICATION'), b_expr
    assert b_expr.operator == 'b', b_expr
    assert b_expr.arguments is None, b_expr
    # field `c` of `b.c`
    assert c_expr == 'c', c_expr


def test_function_application_dot():
    module_name = 'dot_func_prime'
    source = f'''
        ---- MODULE {module_name} ----
        A == a[x].name
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == module_name, module
    assert len(module.units) == 1, module
    # `A` definition
    a_def, = module.units
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # `A` definiens: `a[x].name`
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'FIELD'), a_definiens
    a_x = a_definiens.expression
    name = a_definiens.name
    assert name == 'name', name
    # `a[x]`
    assert (a_x.symbol.name ==
        'FUNCTION_APPLICATION'), a_x
    assert len(a_x.arguments) == 1, a_x.arguments
    a_expr = a_x.function
    x_expr, = a_x.arguments
    assert_operator_name(a_expr, 'a')
    assert_operator_name(x_expr, 'x')


def test_parsing_of_proofs():
    source = '''
        ---- MODULE theorem_and_proof ----
        THEOREM TRUE
        PROOF
        <1>1. TRUE
        <1> QED
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module
    assert module.name == 'theorem_and_proof', module
    assert len(module.units) == 1, module
    theorem, = module.units
    # `THEOREM`
    assert (theorem.symbol.name ==
        'THEOREM'), theorem
    assert theorem.name is None, theorem
    # goal of theorem
    goal = theorem.goal
    assert goal.symbol.name == 'BOOLEAN_LITERAL', goal
    assert goal.value == 'TRUE', goal
    # `PROOF`
    proof = theorem.proof
    assert proof.symbol.name == 'PROOF', proof
    assert len(proof.steps) == 2, proof
    step, qed_step = proof.steps
    # step `<1>1`
    assert (step.symbol.name ==
        'PROOF_STEP'), step
    assert step.proof is None, step
    # step number `<1>1.`
    assert step.name == '<1>1.', step.name
    # expression `TRUE` in step `<1>1`
    main = step.main
    # `QED` step
    assert (qed_step.symbol.name ==
        'PROOF_STEP'), qed_step
    assert qed_step.main == 'QED', qed_step
    assert qed_step.proof is None, qed_step
    # step number `<1>` of `QED` step
    assert qed_step.name == '<1>', qed_step
    # theorem with a leaf proof
    source = '''
        ---- MODULE theorem_and_proof ----
        THEOREM TRUE
        PROOF
        <1>1. TRUE
        <1> QED
            BY <1>1
        ====
        '''
    module = _lr.parse(source)
    theorem, = module.units
    proof = theorem.proof
    _, qed_step = proof.steps
    assert (qed_step.symbol.name ==
        'PROOF_STEP'), qed_step
    assert qed_step.main == 'QED', qed_step
    assert qed_step.proof is not None, qed_step
    # `BY`
    by = qed_step.proof
    assert (by.symbol.name == 'BY'), by
    assert by.names is None, by
    # expressions of `BY`
    facts = by.facts
    assert len(facts) == 1, facts
    # fact `<1>1` in the leaf proof `BY <1>1`
    fact, = facts
    assert (fact.symbol.name ==
        'OPERATOR_APPLICATION'), fact
    assert fact.operator == '<1>1', fact
    assert fact.arguments is None, fact


def test_record_fields():
    source = '''
        ---- MODULE record_fields ----
        A == LET b == FALSE IN b.c
        ====
        '''
    module = _lr.parse(source)
    assert module.symbol.name == 'MODULE', module.symbol
    assert module.name == 'record_fields', module.name
    assert len(module.units) == 1, module
    # `A` definition
    a_def, = module.units
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    assert a_def.name == 'A', a_def
    assert a_def.arity is None, a_def
    assert a_def.local is None, a_def
    # definiens of `A`
    a_definiens = a_def.definiens
    assert (a_definiens.symbol.name ==
        'LET'), a_definiens
    # definitions of `LET` expression
    definitions = a_definiens.definitions
    assert len(definitions) == 1, definitions
    # `b` definition
    b_def, = definitions
    assert (b_def.symbol.name ==
        'OPERATOR_DEFINITION'), b_def
    assert b_def.arity is None, b_def
    assert b_def.local is None, b_def
    # definiens of `b`
    b_definiens = b_def.definiens
    assert (b_definiens.symbol.name ==
        'BOOLEAN_LITERAL'), b_definiens
    assert b_definiens.value == 'FALSE', b_definiens
    # `IN` expression of `LET`
    b_c = a_definiens.expression
    assert (b_c.symbol.name ==
        'FIELD'), b_c
    assert b_c.name == 'c', b_c
    # expression `b`
    b_expr = b_c.expression
    assert (b_expr.symbol.name ==
        'OPERATOR_APPLICATION'), b_expr
    assert b_expr.operator == 'b', b_expr
    assert b_expr.arguments is None, b_expr
    # multiple dots
    parser = _lr.Parser()
    source = '''
        ---- MODULE record_fields ----
        A == a.b.c
        ====
        '''
    module = parser.parse(source)
    assert len(module.units) == 1, module
    a_def, = module.units
    assert (a_def.symbol.name ==
        'OPERATOR_DEFINITION'), a_def
    # expression `a.b.c`
    a_b_c = a_def.definiens
    assert (a_b_c.symbol.name == 'FIELD'), a_b_c
    assert a_b_c.name == 'c', a_b_c
    # expression `a.b`
    a_b = a_b_c.expression
    assert (a_b.symbol.name == 'FIELD'), a_b
    assert a_b.name == 'b', a_b
    # expression `a`
    a_expr = a_b.expression
    assert (a_expr.symbol.name ==
        'OPERATOR_APPLICATION'), a_expr
    assert a_expr.operator =='a', a_expr
    assert a_expr.arguments is None, a_expr


def test_neg_operator_precedence():
    source = r'''
        ---- MODULE experiment ----
        A == TRUE # ~ TRUE
        ====
        '''
    module = _lr.parse(source)
    assert module is not None
    assert len(module.units) == 1, module
    a_def, = module.units
    assert_op_def(a_def, 'A')
    expr = a_def.definiens
    assert_binary_op_app(expr, '#')
    true, neg_true = expr.arguments
    assert_boolean_literal(true, 'TRUE')
    assert_unary_op_app(neg_true, '~')
    true, = neg_true.arguments
    assert_boolean_literal(true, 'TRUE')


def test_unary_minus_operator_precedence():
    source = r'''
        ---- MODULE experiment ----
        A == 1 * - 2
        ====
        '''
    module = _lr.parse(source)
    assert module is not None
    assert len(module.units) == 1, module
    a_def, = module.units
    assert_op_def(a_def, 'A')
    expr = a_def.definiens
    assert_binary_op_app(expr, '*')
    one, minus_two = expr.arguments
    assert_integer_numeral(one, '1')
    assert_unary_op_app(minus_two, '-')
    two, = minus_two.arguments
    assert_integer_numeral(two, '2')
    # unary minus has precedence 12
    # binary plus has precedence 10
    source = '-x + y'
    expr = _lr.parse_expr(source)
    assert expr is not None
    assert_binary_op_app(expr, '+')
    minus_x, y = expr.arguments
    assert_unary_op_app(minus_x, '-')
    x, = minus_x.arguments
    assert_operator_name(x, 'x')
    assert_operator_name(y, 'y')
    # unary minus has precedence 12
    # binary minus has precedence 11
    source = '-x - y'
    expr = _lr.parse_expr(source)
    assert expr is not None
    assert_binary_op_app(expr, '-')
    minus_x, y = expr.arguments
    assert_unary_op_app(minus_x, '-')
    x, = minus_x.arguments
    assert_operator_name(x, 'x')
    assert_operator_name(y, 'y')
    # unary minus has precedence 12
    # binary * has precedence 13
    source = '-x * y'
    expr = _lr.parse_expr(source)
    assert expr is not None
    assert_unary_op_app(expr, '-')
    prod, = expr.arguments
    assert_binary_op_app(prod, '*')
    x, y = prod.arguments
    assert_operator_name(x, 'x')
    assert_operator_name(y, 'y')


def test_prime_function_application():
    text = " f[x]' "
    expr = _lr.parse_expr(text)
    assert expr is not None
    assert_unary_op_app(expr, "'")
    fx, = expr.arguments
    assert (fx.symbol.name ==
        'FUNCTION_APPLICATION'), fx
    assert len(fx.arguments) == 1, fx.arguments
    f_expr = fx.function
    x_expr, = fx.arguments
    assert_operator_name(f_expr, 'f')
    assert_operator_name(x_expr, 'x')
    text = " f'[x] "
    expr = _lr.parse_expr(text)
    assert (expr.symbol.name ==
        'FUNCTION_APPLICATION'), expr
    assert len(expr.arguments) == 1, expr.arguments
    f_primed = expr.function
    x_expr, = expr.arguments
    assert_operator_name(x_expr, 'x')
    assert_unary_op_app(f_primed, "'")
    f_expr, = f_primed.arguments
    assert_operator_name(f_expr, 'f')


def assert_op_def(
        op_def,
        operator_name:
            str):
    """Assert definition of operator,

    without parameters.
    """
    assert op_def.symbol == 'operator_definition', (
        op_def.symbol)
    assert op_def.name == operator_name, op_def.name
    assert op_def.arity is None, op_def.arity
    assert op_def.local is None, op_def.local
    assert op_def.function is None, op_def.function


def assert_unary_op_app(
        expr,
        operator_name:
            str):
    """Assert unary operator application."""
    assert expr.symbol == 'operator_application', (
        expr.symbol)
    assert expr.operator == operator_name, expr.operator
    assert len(expr.arguments) == 1, expr.arguments


def assert_binary_op_app(
        expr,
        operator_name:
            str):
    """Assert binary operator application."""
    assert expr.symbol == 'operator_application', (
        expr.symbol)
    assert expr.operator == operator_name, expr.operator
    assert len(expr.arguments) == 2, expr.arguments


def assert_operator_name(
        expr,
        operator_name:
            str):
    """Assert nullary operator application."""
    assert expr.symbol == 'operator_application', (
        expr.symbol)
    assert expr.operator == operator_name, expr.operator
    assert expr.arguments is None, expr.arguments


def assert_boolean_literal(
        expr,
        value:
            str):
    """Assert Boolean literal with `value`."""
    assert expr.symbol == 'boolean_literal', expr.symbol
    assert expr.value == value, expr.value


def assert_integer_numeral(
        expr,
        value:
            str):
    """Assert integral numeral with `value`."""
    assert expr.symbol == 'integral_numeral', expr.symbol
    assert expr.value == value, expr.value


if __name__ == '__main__':
    # test_preamble_comments()
    # test_vertical_operators()
    test_set_of_tuples()
    # test_unary_minus_as_argument()
    # test_set_theory_operators()
    # test_keyword_unary_operators()
    # test_unary_minus_as_operator()
    # test_set_slice_operator()
    # test_set_comprehension_operator()
    # test_quantifier_vertical_expression()
    # test_vertical_expr_proof_step()
    # test_quantifier_vertical_expression()
    # test_named_theorem()
    # test_quantifier_temporal()
    # test_unary_minus_operator_definition()
    # test_binary_operator_definition()
    # test_nested_proofs()
    # test_pipe_instantiation()
    # test_lambda_inside_undelimited()
    # test_eventually_always()
    # test_let_expression()
    # test_definition_inside_vertical_operator()
    # test_optional_proof_keyword_define()
    # test_dot_function_application_prime()
    # test_parsing_of_proofs()
    # test_record_fields()
    print('tests completed')



"""LR machine that creates RVERTICAL, END_LET as needed."""
import argparse as _arg
import collections as _cl
import collections.abc as _abc
import functools as _ft
import os
import typing as _ty

import parstools._grammars as _grm
import parstools._introspection as _intro
import parstools._lex as _prs_lex
import parstools._lr.parser as _p
from parstools._lr.parser import (
    Result,
    CachedIterable,
    _ROOT,
    _END,
    _State,
    _Actions,
    _TreeMap)
import parstools._lr.utils as _lu

import tla._ast as _tla_ast
import tla._grammar as _tla_grm
import tla._langdef as _opdef
import tla._lex as _lex
import tla._preparser as _prp
import tla._utils as _utils


# auto-generated tokens by the parser,
# with effect similar to the LR end-marker
_AUTO_VIRTUAL_TOKENS: _ty.Final = {
    # 'VERTICAL_OPERATOR',
    # 'LVERTICAL',
    'RVERTICAL',
    # 'DEF_SIG',
    # 'EXCLAMATION_AT',
    # 'EXCLAMATION_COLON',
    # 'EXCLAMATION_DOUBLE_LANGLE',
    # 'EXCLAMATION_DOUBLE_RANGLE',
    # 'EXCLAMATION_NAME',
    # 'END_DEFINE',
    'END_BSLASH_A',
    'END_BSLASH_E',
    'END_BSLASH_AA',
    'END_BSLASH_EE',
    'END_CHOOSE',
    'END_LAMBDA',
    'END_IF',
    'END_CASE',
    'END_LET',
    'END_PICK',
    }
_TOKEN_PAIRS: _ty.Final = dict(
    LVERTICAL='RVERTICAL',
    LET='END_LET',
    BSLASH_A='END_BSLASH_A',
    BSLASH_E='END_BSLASH_E',
    BSLASH_AA='END_BSLASH_AA',
    BSLASH_EE='END_BSLASH_EE',
    CHOOSE='END_CHOOSE',
    LAMBDA='END_LAMBDA',
    IF='END_IF',
    CASE='END_CASE',
    PICK='END_PICK')
_UNDELIMITED_TOKENS = set(_TOKEN_PAIRS)
_UNDELIMITED_TOKENS.remove('LVERTICAL')
_LEXEME_PRECEDENCE = _opdef.LEXEME_PRECEDENCE
_TLA_OPERATOR_FIXITY = _opdef.TLA_OPERATOR_FIXITY
_EXCEPT_TOKEN_TYPES = _tla_grm.EXCEPT_TOKEN_TYPES
_VERTICAL_LEXEMES = _opdef.VERTICAL_LEXEMES


# If the token that precedes an operator
# is in the set `_LOOKBEHIND_TOKENS`,
# then the operator is vertical.
_LOOKBEHIND_TOKENS = {
    # The method
    # `OperatorLexer.__init__()`
    # adds infix and prefix operators
    # to this set.
    'ASSUME',
    'ASSUMPTION',
    'AXIOM',
    'BSLASH_IN',
    'BY',
    'COROLLARY',
    'CARTESIAN',
    'CASE',
    'COLON',
    'COMMA',
    'DOMAIN',
    'DOUBLE_EQ',
    'EQ',
    'ELIF',
    'ELSE',
    'ENABLED',
    'HAVE',
    'HIDE',
    'IF',
    'IN',
    'LBRACE',
    'LBRACKET',
    'DOUBLE_LANGLE',
    'LARROW',
    'LBRACKET_RBRACKET',
    'LEMMA',
    'LPAREN',
    'LVERTICAL',
    'MAPSTO',
    'ONLY',
    'OTHER',
    'PROPOSITION',
    'PROVE',
    'DOUBLE_RANGLE_UNDERSCORE',
    'RARROW',
    'RBRACKET_UNDERSCORE',
    # 'RVERTICAL',
    'SUBSET',
    'SUFFICES',
    'THEN',
    'THEOREM',
    'UNCHANGED',
    'UNION',
    'USE',
    'WF_',
    'WITNESS',
    'SF_',
    'STEP_NUMBER',
    'STEP_NUMBER_ASTERISK',
    'STEP_NUMBER_PLUS',
    }
_utils.assert_len(
    _LOOKBEHIND_TOKENS, 49)


class Grammar(
        _tla_grm.OperatorPrecedenceGrammar):
    """Stacks for undelimited constructs."""

    def __init__(
            self
            ) -> None:
        """Initialize state."""
        super().__init__()
        self.scopes_stack = list()
            # LET, LVERTICAL, quantification,
            # and other undelimited constructs
        self.columns_stack = list()
            # opening columns of
            # vertical operators

    def p_vertical_operator(
            self, p):
        """vertical_op: VERTICAL_OPERATOR"""
        # `_map_if_vertical()` creates
        # `VERTICAL_OPERATOR` tokens
        p[0] = p[1]
        self.scopes_stack.append(
            'LVERTICAL')
        trees = p[-1]
        if len(trees) != 1:
            raise AssertionError(trees)
        vertical_op, = trees
        self.columns_stack.append(
            vertical_op.column)
        # The LR parser via this method
        # changes the stacks only when
        # a `VERTICAL_OPERATOR` occurrence
        # that starts a vertical item.


class OperatorLexer:
    """Adapts token sequence for parser."""

    def __init__(
            self
            ) -> None:
        """Initialization."""
        self._lexer = _lex.Lexer()
        self._token_grouping, before_vertical = (
            self._make_token_types())
        self._before_vertical = before_vertical

    def _make_token_types(
            self
            ) -> dict:
        """Make grouping of tokens."""
        keywords = {
            v: v for v in _opdef.KEYWORDS}
        lexemes_to_tokens = (
            keywords |
            _lex.map_lexemes_to_tokens())
        before_vertical = set(_LOOKBEHIND_TOKENS)
        groups = dict()
        items = _LEXEME_PRECEDENCE.items()
        for lexeme, prec in items:
            old_type = lexemes_to_tokens[lexeme]
            if old_type in _EXCEPT_TOKEN_TYPES:
                continue
            new_type = self._type_of_prec(
                lexeme, prec)
            groups[old_type] = new_type
            self._update_lookbehind_token_types(
                new_type, before_vertical)
        return (
            groups,
            before_vertical)

    def _type_of_prec(
            self,
            lexeme:
                str,
            prec:
                _opdef.PrecRange
            ) -> str:
        """Return token type for precedence."""
        start, _ = prec
        items = _TLA_OPERATOR_FIXITY.items()
        for fixity, operators in items:
            if lexeme not in operators:
                continue
            return f'OP_{start}_{fixity}'.upper()

    _PREFIX_OR_INFIX: _ty.Final = {
        'PREFIX',
        'BEFORE',
        'INFIX',
        'LEFT',
        'BETWEEN',
        'RIGHT'}

    def _update_lookbehind_token_types(
            self,
            new_type,
            before_vertical):
        """Add prefix and infix operators to

        `before_vertical`.
        """
        if new_type is None:
            return
        *_, fixity = new_type.rpartition('_')
        if fixity in ('POSTFIX', 'AFTER'):
            return
        if fixity in self._PREFIX_OR_INFIX:
            before_vertical.add(new_type)
            return
        raise ValueError(
            new_type)

    def _group_tokens(
            self,
            tokens
            ) -> _abc.Iterable[
                _prs_lex.Token]:
        """Adapt token symbols."""
        group = self._token_grouping.get
        new_tokens = list()
        for token in tokens:
            ts = token.symbol
            new_type = group(ts, ts)
            new_token = _utils.copy_lex_token(
                token,
                symbol=new_type)
            new_tokens.append(new_token)
        return new_tokens

    def parse(
            self,
            text:
                str
            ) -> _abc.Iterable[
                _prs_lex.Token]:
        """Return tokenized `text`."""
        tokens = self._lexer.parse(text)
        tokens = _prp.filter_comments(tokens)
        tokens = _prp.filter_outside_module(tokens)
        tokens = self._group_tokens(tokens)
        tokens = _prp.preparse(tokens)
        return tokens


class GeneratingParser:
    """Shift-reduce parser.

    For a set of "virtual tokens",
    the parser generates them when
    they are missing from the input.

    This approach is correct for
    handling syntactic features of
    a number of languages.

    Actions are structured as:
    `dict[node, dict[lookahead, ...]]`
    """

    def __init__(
            self
            ) -> None:
        """Initialize."""
        # cache filename
        filename = 'tla_parser.json'
        path, _ = os.path.split(__file__)
        cache_filepath = os.path.join(
            path, filename)
        # LR tables
        self.grammar = Grammar()
        op_prec = self.grammar.operator_precedence
        parser = _intro.make_parser(
            self.grammar,
            operator_precedence=op_prec,
            cache_filepath=cache_filepath)
        self._initial: _State = parser._initial
        actions = self._restructure_actions(
            parser._actions)
        self._actions: _Actions = actions
        self._tree_map: _TreeMap = parser._tree_map
        # stacks
        lexer = OperatorLexer()
        self._before_vertical = lexer._before_vertical
        self._last_token = None

    def _restructure_actions(
            self,
            actions:
                dict
            ) -> dict:
        """Return nested dictionary."""
        actions_ns = _cl.defaultdict(dict)
        kvs = actions.items()
        for (node, symbol), action in kvs:
            actions_ns[
                node][symbol] = action
        return dict(actions_ns)

    def _clear_state(
            self
            ) -> None:
        """Initialize parser variables."""
        self._last_token = None
        self.grammar.scopes_stack.clear()
        self.grammar.columns_stack.clear()

    def parse(
            self,
            items:
                _abc.Iterable[
                    Result]
            ) -> Result | None:
        """Reduce items to a tree."""
        self._clear_state()
        sequence = list(items)
        end = Result(
            symbol=_END,
            value=_END)
        sequence.append(end)
        symbols = CachedIterable(sequence)
            # input stack
        results = list()
            # output stack
        path = [self._initial]
        shift_or_reduce = _ft.partial(
            self._shift_or_reduce,
            symbols=symbols,
            results=results,
            path=path)
        while symbols.peek() and path:
            self._map_if_vertical(symbols, results)
            self._generate_dedents(symbols)
            shift_or_reduce()
        if symbols.peek() or path:
            return None
        if len(results) != 1:
            raise AssertionError(results)
        if self.grammar.scopes_stack:
            raise AssertionError(
                self.grammar.scopes_stack)
        if self.grammar.columns_stack:
            raise AssertionError(
                self.grammar.columns_stack)
        return results[0]

    def _shift_or_reduce(
            self,
            symbols:
                list[
                    Result],
            results:
                list[
                    Result],
            path:
                list
            ) -> None:
        """Step of shift-reduce parser."""
        if not path:
            raise AssertionError(path)
        peek = symbols.peek()
        if peek is None:
            raise AssertionError(symbols)
        lookahead = peek.symbol
        # if isinstance(peek.value, str):
        #     print(peek)
        if lookahead == _ROOT:
            results.append(symbols.pop())
            treelet = symbols.pop()
            if treelet.symbol != _END:
                raise AssertionError(_END)
            path.pop()
            return
        node = path[-1]
        action = self._actions[node].get(lookahead)
        # generate virtual token ?
        generate = (
            action is None and
            self.grammar.scopes_stack)
        if generate:
            self._generate_virtual_tokens(
                node, symbols, results)
            return
        # parsing error ?
        if action is None:
            self._print_info(symbols, results)
            path.clear()
            return
        if action.equation is not None:
            eq = action.equation
            symbol, expansion = eq
            n_items = len(expansion)
            _p.popk(path, n_items)
            equation = str(eq)
            trees = _p.popk(results, n_items)
            if self._tree_map is not None:
                reduce = self._tree_map[equation]
                p = [None, *(
                    u.value
                    for u in trees),
                    trees]
                reduce(p)
                trees = p[0]
            result = Result(
                symbol=symbol,
                value=trees,
                equation=equation)
            symbols.append(result)
        elif action.state is not None:
            path.append(action.state)
            treelet = symbols.pop()
            results.append(treelet)
            self._scopes_stack_append(
                treelet.symbol)
            # leaf ?
            if isinstance(treelet.value, str):
                self._last_token = treelet
        else:
            raise AssertionError(
                node, lookahead, action)

    def _scopes_stack_append(
            self,
            symbol:
                str
            ) -> None:
        """Append if start of undelimited."""
        if symbol not in _UNDELIMITED_TOKENS:
            return
        self.grammar.scopes_stack.append(symbol)

    def _map_if_vertical(
            self,
            symbols:
                CachedIterable,
            results:
                list
            ) -> None:
        """Map tokens to vertical tokens.

        Create `VERTICAL_OPERATOR` tokens from
        infix operators in `_VERTICAL_LEXEMES`,
        if the occurrences are vertical.
        """
        # NOTE: an LR shift-reduce overlap
        # gave rise to this function
        token = symbols.peek()
        if token is None:
            return
        # nonleaf ?
        if not isinstance(token.value, str):
            return
        is_vertical = (
            token.symbol != 'VERTICAL_OPERATOR' and
                # avoid recursion
            token.value in _VERTICAL_LEXEMES and
            ((self._last_token.symbol in
                self._before_vertical) or
                # tokens and not nonleafs
            (self._last_token.symbol == 'RVERTICAL' and
                self._last_token.column == token.column and
                    # same column
                len(results) >= 4 and
                results[-4].symbol == 'vertical_op' and
                results[-4].value == token.value)))
                    # same symbol of vertical operator,
                    # e.g., "/\\"
        if not is_vertical:
            return
        vertical_op = _utils.copy_lex_token(
            token,
            symbol='VERTICAL_OPERATOR')
        symbols.pop()
        lvertical = _prs_lex.Token(
            symbol='LVERTICAL',
            value='LVERTICAL',
            column=token.column + 1,
                # Avoid closing vertical expression
                # due to indentation.
                #
                # Using `token.column` would
                # generate an `RVERTICAL` before
                # shifting the `LVERTICAL`.
            row=token.row,
            start=None)
        symbols.append(lvertical)
        symbols.append(vertical_op)

    def _generate_virtual_tokens(
            self,
            node,
            symbols:
                list,
            results:
                list
            ) -> None:
        """Insert a suitable virtual token.

        Virtual tokens are listed in
        `_AUTO_VIRTUAL_TOKENS`.

        `results` is used only for printing
        information if an error occurs.
        """
        scopes_stack = self.grammar.scopes_stack
        columns_stack = self.grammar.columns_stack
        node_actions = self._actions[node]
        vtokens = _AUTO_VIRTUAL_TOKENS.intersection(
            node_actions)
        stack_token = scopes_stack[-1]
        token_value = _TOKEN_PAIRS[stack_token]
        if token_value not in vtokens:
            print(
                'Expected to insert token:\n'
                f'{token_value}\n'
                'Possible next tokens based on '
                f'parser state machine:\n{vtokens}')
            self._print_info(symbols, results)
            raise ValueError('Error while parsing.')
        if stack_token == 'LVERTICAL':
            if not columns_stack:
                raise AssertionError(
                    scopes_stack)
            column = columns_stack[-1]
        else:
            column = None
        token = Result(
            symbol=token_value,
            value=token_value,
            column=column)
        symbols.append(token)
        self._pop_scope_assert(stack_token)
        # print(f'auto-generated: {token_value}')

    def _pop_scope_assert(
            self,
            symbol:
                str
            ) -> None:
        """Pop expected symbol."""
        scopes_stack = self.grammar.scopes_stack
        token_symbol = scopes_stack[-1]
        if token_symbol != symbol:
            raise AssertionError(
                token_symbol)
        scopes_stack.pop()
        columns_stack = self.grammar.columns_stack
        if symbol == 'LVERTICAL':
            columns_stack.pop()
        inv = (
            len(scopes_stack) >=
            len(columns_stack))
        if inv:
            return
        raise AssertionError(
            scopes_stack,
            columns_stack)

    def _generate_dedents(
            self,
            symbols
            ) -> None:
        """Insert dedents.

        Inserts `RVERTICAL` due to dedent.
        """
        scopes_stack = self.grammar.scopes_stack
        columns_stack = self.grammar.columns_stack
        if not columns_stack:
            return
        if not scopes_stack:
            raise AssertionError(columns_stack)
        # get box column
        box_column = columns_stack[-1]
        if box_column < 0:
            raise AssertionError(box_column)
        # get next-token column
        next_token = symbols.peek()
        not_generate = (
            next_token.symbol == 'RVERTICAL' or
                # CASE: nested vertical expressions,
                # avoid circular recursion
            next_token.column is None or
                # nonleaf next
            next_token.column > box_column)
                # no dedent
        if not_generate:
            return
        # dedent: generate `RVERTICAL`
        dedent = _prs_lex.Token(
            symbol='RVERTICAL',
            value='RVERTICAL',
            column=box_column,
            row=next_token.row,
            start=None)
        index = _utils.rindex(
            'LVERTICAL', scopes_stack)
        scopes_stack.pop(index)
        columns_stack.pop()
        symbols.append(dedent)
        # print('RVERTICAL by dedent')
        if len(scopes_stack) < len(columns_stack):
            raise AssertionError(
                len(scopes_stack),
                len(columns_stack))

    def _print_info(
            self,
            symbols,
            results
            ) -> None:
        """Print information."""
        if results:
            result = results[-1]
            print(f'Top of stack `results`:\n{result}')
        else:
            print('Empty stack `results`')
        next_symbols = list()
        for _ in range(2):
            if symbols.peek() is None:
                break
            next_symbols.append(symbols.pop())
        print(f'Next symbols in input:\n{next_symbols}')
        print(
            '`scopes_stack`:\n'
            f'{self.grammar.scopes_stack}')
        print(
            '`columns_stack`:\n'
            f'{self.grammar.columns_stack}')


class Parser:
    """Parser of TLA+ modules."""

    def __init__(
            self
            ) -> None:
        self._lexer = None
        self._parser = None

    def parse(
            self,
            module_text:
                str
            ) -> tuple:
        """Return syntax tree from `module_text`."""
        if self._parser is None:
            self._lexer = OperatorLexer()
            self._parser = GeneratingParser()
        tokens = self._lexer.parse(module_text)
        result = self._parser.parse(tokens)
        if result is None:
            raise ValueError('Error while parsing.')
        tree = result.value
        tree = _tla_ast.map_steps_to_proofs(tree)
        return tree


def lex(
        text:
            str
        ) -> list[
            _p.Result]:
    """Parse lexemes."""
    lexer = OperatorLexer()
    tokens = lexer.parse(text)
    tokens = list(tokens)
    # for token in tokens:
    #     print(token)
    return tokens


class ExprParser(
        Parser):
    """Parser of TLA+ expressions."""

    def parse(
            self,
            tla_expr:
                str
            ) -> _ty.Any:
        """Parse a TLA+ expression."""
        tla_module_text = f'''
            ---- MODULE name ----
            AXIOM name == \n{tla_expr}
            ====
            '''
        module_tree = super().parse(
            tla_module_text)
        expr_tree = module_tree.units[0].expression
        return expr_tree


def parse(
        module_text:
            str
        ) -> tuple:
    """Return syntax tree for `module_text`."""
    if not hasattr(parse, 'parser'):
        parse.parser = Parser()
    return parse.parser.parse(module_text)


def parse_expr(
        expr_text:
            str
        ) -> tuple:
    """Return syntax tree for `expr_text`."""
    if not hasattr(parse_expr, 'parser'):
        parse_expr.parser = ExprParser()
    return parse_expr.parser.parse(expr_text)


def _parse_file(
        filename:
            str
        ) -> None:
    """Parsing file `filename`."""
    with open(filename, 'r') as fd:
        file_text = fd.read()
    tree = parse(file_text)
    print(tree)
    _tla_ast.pprint_tree(tree)


def _main():
    """Entry point."""
    args = _parse_args()
    filename = args.filename
    _parse_file(filename)


def _parse_args(
        ) -> _arg.Namespace:
    """Return arguments."""
    parser = _arg.ArgumentParser()
    parser.add_argument(
        'filename',
        help='TLA+ file to parse')
    return parser.parse_args()


if __name__ == '__main__':
    _main()



"""Count nonempty, comment, blank lines in a TLA+ file."""
import argparse as _arg

import tla


_COMMENT_TOKENS = {
    'UNILINE_COMMENT',
    'MULTILINE_COMMENT'}


def count_lines(
        text:
            str
        ) -> dict[
            str, int]:
    """Return count of nonempty, comment, blank lines."""
    lexer = tla.Lexer()
    tokens = list(lexer.parse(text))
    comment_rows = set()
    nonempty_rows = set()
    other_rows = set()
    for token in tokens:
        start_row = token.row
        nonempty_rows.add(start_row)
        if token.symbol not in _COMMENT_TOKENS:
            other_rows.add(start_row)
            continue
        if token.symbol == 'UNILINE_COMMENT':
            comment_rows.add(start_row)
        if token.symbol != 'MULTILINE_COMMENT':
            raise AssertionError(token)
        n_newlines = token.value.count('\n')
        for diff in range(n_newlines + 1):
            row = start_row + diff
            comment_rows.add(row)
            nonempty_rows.add(row)
    n_nonempty = len(nonempty_rows)
    only_comment = comment_rows - other_rows
    n_comments = len(only_comment)
    n_lines = text.count('\n')
    if not text.endswith('\n'):
        n_lines += 1
    n_blank = n_lines - n_nonempty
    n_other = n_nonempty - len(only_comment)
    return dict(
        n_lines=n_lines,
        n_comments=n_comments,
        n_blank=n_blank,
        n_other=n_other)


def _main():
    """Entry point."""
    args = _parse_args()
    filename = args.filename
    with open(filename, 'r') as fd:
        text = fd.read()
    line_stats = count_lines(text)
    lines = '\n'.join(
        f'{k} = {v}'
        for k, v in line_stats.items())
    print(lines)


def _parse_args(
        ) -> _arg.Namespace:
    """Return program arguments."""
    parser = _arg.ArgumentParser(
        description='Line statistics for TLA+ files.')
    parser.add_argument(
        'filename',
        type=str,
        help='A TLA+ file.')
    return parser.parse_args()


if __name__ == '__main__':
    _main()



"""TLA+ parser and syntax tree."""
try:
    import tla._version as _version
    __version__ = _version.version
except ImportError:
    __version__ = None

import tla._ast as _ast
import tla._lex as _lex
import tla._lre as _lr
import tla._pprint as _pp


Lexer = _lex.Lexer
Parser = _lr.Parser
ExprParser = _lr.ExprParser
parse = _lr.parse
parse_expr = _lr.parse_expr
pformat_ast = _pp.pformat_ast
pformat_tla = _pp.pformat_tla
make_nodes = _ast.make_nodes



"""Parse the files in the TLAPS library."""
import os

import tla._lre as _parser
import tla._pprint as _pp_tla


# change this variable to a path where
# the TLAPS library is present
TLAPS_LIB_PATH = '$HOME/git/tlapm/library'


def parse_tlaps_modules():
    module_paths = _collect_tlaps_module_files()
    parser = _parser.Parser()
    for module_path in module_paths:
        print(f'parsing module `{module_path}`')
        text = _read_file(module_path)
        _parse_and_format(text, parser)


def _collect_tlaps_module_files(
        ) -> list[
            str]:
    path = TLAPS_LIB_PATH
    path = os.path.expandvars(path)
    tla_files = list()
    with os.scandir(path) as it:
        for entry in it:
            is_tla_file = (
                entry.is_file() and
                entry.name.endswith('.tla'))
            if is_tla_file:
                tla_files.append(entry)
    return [
        entry.path
        for entry in tla_files]


def _read_file(
        path:
            str
        ) -> str:
    with open(path, 'r') as fd:
        return fd.read()


def _parse_and_format(
        text:
            str,
        parser):
    """Return syntax-tree from `text`."""
    tree = parser.parse(text)
    if tree is None:
        raise AssertionError(r)
    text = _pp_tla.pformat_ast(tree)
    print(text)
    # _pp_tla._print_overwide_lines(
    #     text, _pp_tla.LINE_WIDTH)
    return tree


if __name__ == '__main__':
    parse_tlaps_modules()



"""Interpret arithmetic TLA+ expressions."""
import tla


def main():
    """Entry point."""
    expression = '1 + 2 * 3.5'
    number = calculate(expression)
    print(number)


def calculate(
        expression:
            str
        ) -> float:
    """Return the numeric value of `expression`."""
    tree = tla.parse_expr(expression)
    return evaluate(tree)


def evaluate(
        tree
        ) -> float:
    match tree.symbol:
        case 'operator_application':
            args = list(map(
                evaluate,
                tree.arguments))
            match tree.operator:
                case '+':
                    assert len(args) == 2, args
                    return args[0] + args[1]
                case '-':
                    # unary minus ?
                    if len(args) == 1:
                        return args[0]
                    assert len(args) == 2, args
                    return args[0] - args[1]
                case '*':
                    assert len(args) == 2, args
                    return args[0] * args[1]
                case '/':
                    assert len(args) == 2, args
                    return args[0] / args[1]
                case op:
                    raise NotImplementedError(
                        f'operator `{op}`')
        case 'integral_numeral':
            return int(tree.value)
        case 'float_numeral':
            return float(tree.value)
        case symbol:
            raise ValueError(symbol)


if __name__ == '__main__':
    main()



"""Tests for the package `tla`."""
import pytest
import tla._langdef as _opdef
import tla._lex as _lex
import tla._lre as _lr


def test_check_lexprec():
    _opdef.check_lexprec()


expr_tests = [
    " FALSE ",
    " TRUE ",
    " ~ FALSE ",
    " ~ TRUE ",
    " ~ x ",
    " ~ y ",
    " ~ 1 ",
    " ~ 2 ",
    " FALSE /\\ TRUE ",
    " x /\\ y ",
    " FALSE \\/ TRUE ",
    " x \\/ y ",
    " FALSE => TRUE ",
    " x => y ",
    " FALSE <=> TRUE ",
    " x <=> y ",
    " IF FALSE THEN FALSE ELSE FALSE ",
    " IF x = 1 THEN 2 ELSE 3 ",
    " x = 5 ",
    " y = 2 ",
    " x # 5 ",
    " y # 2 ",
    " Nat ",
    " 15 ",
    ' "15" ',
    " @ ",
    " x @@ y ",
    " (x @@ y) ",
    " x :> y ",
    " (x :> y) ",
    " x <: y ",
    " (x <: y) ",
    " x + y ",
    " (x + y) ",
    " x - y ",
    " (x - y) ",
    " - x ",
    " (- x) ",
    " -x ",
    " (-x) ",
    " - 1 ",
    " (- 1) ",
    " -1 ",
    " (-1) ",
    " x / y ",
    " (x / y) ",
    " x % y ",
    " (x % y) ",
    " (x + 1) ",
    # quantification
    " \\A y \\in R:  \\A <<u, v>> \\in S:  G(y, u, v) ",
    " \\E x, y:  x + y < 0 ",
    " \\AA x: TRUE ",
    " \\EE x, y: TRUE ",
    # `CHOOSE`
    " CHOOSE foo: FALSE ",
    r" CHOOSE <<u, v>> \in S:  A(u, v) ",
    # set theory
    " {1, 2, 3, 4 / 5} ",
    " {Foo \\in FALSE: FALSE} ",
    " {FALSE: x \\in {1, 2}} ",
    " {FALSE, FALSE} ",
    " SUBSET A.r ",
    " UNION {S \\in A:  S \\cap Q # {}} ",
    " {x \\in Int:  x < 0} ",
    " x \\subseteq y ",
    " x \\in S ",
    " x \\notin S ",
    " x \\in 1 ",
    " x \\in {1, 2} ",
    " x \\notin 2 ",
    " x \\notin {3, 4} ",
    " S \\ Q ",
    " S \\cap Q ",
    " S \\cup Q ",
    " A \\X B \\X C ",
    " "
    # functions
    " [foo |-> FALSE] ",
    " [foo |-> FALSE, bar |-> FALSE] ",
    " [FALSE -> FALSE] ",
    " [x, y \\in S, uv \\in T |-> G(x, y, uv[1], uv[2])] ",
    " DOMAIN f ",
    r" [x, y \in S, <<u, v>> \in T |-> G(x, y, u, v)] ",
    " f[x]' ",
    " f[k + 1, r] ",
    " [(A \\cup B) -> C] ",
    " [a |-> u + v, b |-> w] ",
    " [a: Int, b: A \\cup B] ",
    # EXCEPT
    " [f EXCEPT !.foo = FALSE] ",
    " [f EXCEPT ![FALSE] = FALSE] ",
    " [f EXCEPT ![FALSE] = FALSE, !.foo = FALSE] ",
    "[f EXCEPT ![1, x] = e] ",
    " [f EXCEPT ![1].g = 2, ![<<t, 3>>] = u] ",
    # tuples
    " <<TRUE>> ",
    " <<TRUE, TRUE>> ",
    " <<1, 2, 3 + 5 >> ",
    " Int \\X (1..2) \\X Real ",
    # actions
    " (TRUE + 1)' ",
    " (1 + 2)' /\\ 3 ",
    " [x \\in {1, 2} |-> x' + 1] ",
    " [FALSE]_TRUE ",
    " <<TRUE>>_TRUE ",
    " [x' = x + 1]_<<x, y>> ",
    " <<x' = x + 1>>_<<x, y>> ",
    " [M \\/ N]_<<u, v>> ",
    " <<x' = x + 1>>_(x * y) ",
    " ENABLED (x' = x + 1) ",
    " (x = 0 /\\ x' = 1) \\cdot "
        "(x = 1 /\\ x' = 2) ",
    " UNCHANGED x ",
    " UNCHANGED FALSE ",
    " UNCHANGED TRUE ",
    " UNCHANGED 1 ",
    " UNCHANGED <<x>> ",
    " UNCHANGED <<x, y>> ",
    " UNCHANGED <<x, y, z>> ",
    # temporal formulas
    " []A ",
    " <>A ",
    " [][A]_v ",
    " [][x' = x + 1]_<<x, y>> ",
    " <><<A>>_v ",
    " <><<x' = x + y>>_<<x, y, z>> ",
    " A -+-> B ",
    " ([]p) -+-> ([]q) ",
    # liveness formulas
    " A ~> B ",
    " WF_TRUE (TRUE) ",
    " WF_<<x, y>>(N /\\ M)",
    " WF_x(N) ",
    " SF_TRUE(TRUE) ",
    # `LAMBDA`
    " LAMBDA Foo:  TRUE ",
    # `LET ... IN`
    " LET a == 1 IN a ",
    " LET x == 1  y == 2 IN x + y ",
    " A(y - 1)!B!C ",
    " A!Op(y - 1, B) ",
    # user-defined operators
    " (x ^ y) ",
    " (x / y) ",
    " (x * y) ",
    " (-x) ",
    " (x + y) ",
    " (x^+) ",
    " (x^*) ",
    " (x^#) ",
    " (x < y) ",
    " (x =< y) ",
    " (x <= y) ",
    " (x > y) ",
    " (x >= y) ",
    " (x ... y) ",
    " (x .. y) ",
    " (x | y) ",
    " (x || y) ",
    " (x && y) ",
    " (x & y) ",
    " (x $$ y) ",
    " (x $ y) ",
    " (x ?? y) ",
    " (x %% y) ",
    " (x % y) ",
    " (x ## y) ",
    " (x ++ y) ",
    " (x -- y) ",
    " (x ** y) ",
    " (x // y) ",
    " (x ^^ y) ",
    " (x @@ y) ",
    " (x !! y) ",
    " (x |- y) ",
    " (x |= y) ",
    " (x -| y) ",
    " (x =| y) ",
    " (x <: y) ",
    " (x :> y) ",
    " (x := y) ",
    " (x ::= y) ",
    " (x (+) y) ",
    " (x (-) y) ",
    " (x (.) y) ",
    " (x (/) y) ",
    " (x (\\X) y) ",
    " (x \\uplus y) ",
    " (x \\sqcap y) ",
    " (x \\sqcup y) ",
    " (x \\div y) ",
    " (x \\wr y) ",
    " (x \\star y) ",
    " (x \\o y) ",
    " (x \\bigcirc y) ",
    " (x \\bullet y) ",
    " (x \\prec y) ",
    " (x \\succ y) ",
    " (x \\preceq y) ",
    " (x \\succeq y) ",
    " (x \\sim y) ",
    " (x \\simeq y) ",
    " (x \\ll y) ",
    " (x \\gg y) ",
    " (x \\asymp y) ",
    " (x \\subset y) ",
    " (x \\supseteq y) ",
    " (x \\approx y) ",
    " (x \\cong y) ",
    " (x \\sqsubset y) ",
    " (x \\sqsubseteq y) ",
    " (x \\sqsupset y) ",
    " (x \\sqsupseteq y) ",
    " (x \\doteq y) ",
    " (x \\propto y) ",
    ]

expr_tests.append(r'''
    /\ \/ FALSE
       \/ FALSE
    /\ FALSE
    ''')
expr_tests.append(r'''
    /\ (f = [x \in {1, 2} |-> x' + 1])
    /\ a = b
    ''')
expr_tests.append(r'''
    CASE   A -> u
        [] B -> v
        [] OTHER -> w
    ''')
expr_tests.append(r'''
    LET
        x == y - 1
        f[u \in Int] == u^2
    IN
        x + f[v]
    ''')

# tests for module parser
module_tests = list()
module_tests.append(r'''
    ---- MODULE definitions ----
    a == 1
    a == INSTANCE M WITH x <- 1
    Foo(x, y) == x + y
    Foo(x, Bar(_)) == x + y

    a == /\ f = [x \in {1, 2} |-> x]
         /\ g = b
    ====
    ''')
module_tests.append(r'''
    ---- MODULE sequents ----
    THEOREM
        ASSUME TRUE
        PROVE TRUE

    THEOREM
        ASSUME NEW x \in {1, 2}
        PROVE x \in {1, 2, 3}
    ====
    ''')
module_tests.append(r'''
    ---- MODULE Foo ----
    x == 1
    ====================
    ''')
module_tests.append(r'''
    ---- MODULE Bar ----
    b == /\ a = 1
         /\ c = d
    ====================
    ''')
module_tests.append(r'''
---- MODULE Foo_Bar ----
EXTENDS Foo

VARIABLE x

Foo == TRUE  (*{ by (prover:"smt3") }*)
Bar == f[<<a, b>>]
OpLabel == label(b):: 1


THEOREM
    \E r, q:  TRUE
PROOF
<1>1. TRUE
    OBVIOUS
<1>2. TRUE
    OMITTED
<1> USE ONLY TRUE, TRUE DEF Foo
<1> USE TRUE
<1> HIDE <1>2
<1> DEFINE r == 1
<1>3. ASSUME TRUE
      PROVE TRUE
<1> QED

------------------------

THEOREM
    ASSUME TRUE, 1, NEW r \in S,
        ASSUME TRUE
        PROVE TRUE
    PROVE TRUE
PROOF
<1>1. TRUE
    OBVIOUS
<1> SUFFICES TRUE
<1> TAKE r \in S, q \in S
<1> QED


(* A multi-line comment.

*)

(*
(* A nested multi-line comment.
*)
*)
========================
''')


def test_expr_parser():
    """Testing of expression parser."""
    lexer = _lex.Lexer()
    parser = _lr.ExprParser()
    for expr in expr_tests:
        tokens = list(lexer.parse(expr))
        print(tokens)
        print(' ')
        print(expr)
        r = parser.parse(expr)
        assert r is not None
        print(r)


def test_module_parser():
    """Testing of module parser."""
    parser = _lr.Parser()
    for module in module_tests:
        print(module)
        r = parser.parse(module)
        assert r is not None


def test_parser_stacks_cleared():
    """Ensure parser initializes its state."""
    parser = _lr.Parser()
    # input without errors
    text = '---- MODULE name ---- A == 1 ===='
    module = parser.parse(text)
    assert module is not None, module
    assert module.symbol == 'module', module
    assert not parser._parser.grammar.scopes_stack, (
        parser._parser.grammar.scopes_stack)
    assert not parser._parser.grammar.columns_stack, (
        parser._parser.grammar.columns_stack)
    # syntax error in input
    text = '---- MODULE name ---- A == LET x == 1 IN ===='
    with pytest.raises(ValueError):
        parser.parse(text)
        # the stacks are nonempty due to the error,
        # but will be cleared when `parse()` is
        # called again
    # again input without errors
    text = '''---- MODULE name ---- A ==
        LET x == 1 IN x ===='''
    module = parser.parse(text)
    assert module is not None, module
    assert module.symbol == 'module', module
    assert not parser._parser.grammar.scopes_stack, (
        parser._parser.grammar.scopes_stack)
    assert not parser._parser.grammar.columns_stack, (
        parser._parser.grammar.columns_stack)


if __name__ == '__main__':
    test_check_lexprec()
    test_expr_parser()
    test_module_parser()



"""Tests of module `tla._pprint`."""
import textwrap as _tw

import pytest
import tla._ast
import tla._lre as _lr
import tla._pprint as _pp


nodes = tla._ast.make_nodes()
pformat = _pp.pformat_ast


def test_axiom():
    # named AXIOM
    expression = _boolean_expr()
    axiom = nodes.Axiom(
        name='name',
        expression=expression)
    text = pformat(axiom)
    text_ = (
        'AXIOM name ==\n'
        '    TRUE')
    assert text == text_, text
    # unnamed AXIOM
    axiom = nodes.Axiom(
        expression=expression)
    text = pformat(axiom)
    text_ = (
        'AXIOM\n'
        '    TRUE')
    assert text == text_, text
    # named ASSUMPTION
    axiom = nodes.Axiom(
        name='true',
        expression=expression,
        axiom_keyword='ASSUMPTION')
    text = pformat(axiom)
    text_ = (
        'ASSUMPTION true ==\n'
        '    TRUE')
    assert text == text_, text
    # named ASSUME
    axiom = nodes.Axiom(
        name='true',
        expression=expression,
        axiom_keyword='ASSUME')
    text = pformat(axiom)
    text_ = (
        'ASSUME true ==\n'
        '    TRUE')
    assert text == text_, text
    # other
    axiom = nodes.Axiom(
        name='axiom_name',
        expression=expression,
        axiom_keyword='AXIOMS')
    with pytest.raises(ValueError):
        pformat(axiom)


def test_boolean_literal():
    # FALSE
    false = nodes.BooleanLiteral(
        value='FALSE')
    text = pformat(false)
    text_ = 'FALSE'
    assert text == text_, text
    # TRUE
    true = nodes.BooleanLiteral(
        value='TRUE')
    text = pformat(true)
    text_ = 'TRUE'
    assert text == text_, text
    # other
    other = nodes.BooleanLiteral(
        value='TRUES')
    with pytest.raises(ValueError):
        pformat(other)


def test_by():
    indent = 4 * '\x20'
    # BY ONLY
    expr = _x_eq_one()
    op_name = nodes.OperatorApplication(
        operator='operator_name')
    facts = [expr]
    names = [op_name]
    by = nodes.By(
        only=True,
        facts=facts,
        names=names)
    text = pformat(by)
    text_ = f'{indent}BY ONLY x = 1 DEF operator_name'
    assert text == text_, text
    # with commas
    facts = [expr, op_name]
    by = nodes.By(
        facts=facts)
    text = pformat(by)
    text_ = f'{indent}BY x = 1, operator_name'
    assert text == text_, text
    # with commas in DEF
    other_op = nodes.OperatorApplication(
        operator='other_op')
    names = [op_name, other_op]
    by = nodes.By(
        names=names)
    text = pformat(by)
    text_ = f'{indent}BY DEF operator_name, other_op'
    assert text == text_, text


def test_cases():
    # 2 cases
    x = _opname('x')
    y = _opname('y')
    u = _opname('u')
    v = _opname('v')
    case_1 = nodes.CaseItem(
        predicate=x,
        expression=u)
    case_2 = nodes.CaseItem(
        predicate=y,
        expression=v)
    cases = [case_1, case_2]
    cases_tree = nodes.Cases(
        cases=cases)
    text = pformat(cases_tree)
    text_ = 'CASE x -> u [] y -> v'
    assert text == text_, text
    # 1 case, OTHER
    expr = _x_eq_one()
    cases = [case_2]
    cases_tree = nodes.Cases(
        cases=cases,
        other=expr)
    text = pformat(cases_tree)
    text_ = 'CASE y -> v [] OTHER -> x = 1'
    assert text == text_, text
    # 3 cases, OTHER
    x_in_s = _x_in_s()
    w = _opname('w')
    case_3 = nodes.CaseItem(
        predicate=x_in_s,
        expression=w)
    cases = [case_1, case_2, case_3]
    cases_tree = nodes.Cases(
        cases=cases,
        other=expr)
    text = pformat(cases_tree)
    text_ = (
        'CASE x -> u [] y -> v [] '
        r'x \in S -> w [] OTHER -> x = 1')
    assert text == text_, text


def test_case_item():
    predicate = _x_eq_one()
    expression = _boolean_expr()
    case_item = nodes.CaseItem(
        predicate=predicate,
        expression=expression)
    text = pformat(case_item)
    text_ = 'x = 1 -> TRUE'
    assert text == text_, text


def test_choose():
    declaration = _x_in_s()
    predicate = _x_eq_one()
    choose = nodes.Choose(
        declaration=declaration,
        predicate=predicate)
    text = pformat(choose)
    text_ = r'CHOOSE x \in S:  x = 1'
    assert text == text_, text


def test_constants():
    x = _opname('x')
    y = _opname('y')
    names = [x, y]
    constants = nodes.Constants(
        names=names)
    text = pformat(constants)
    text_ = 'CONSTANT\n    x,\n    y'
    assert text == text_, text


def test_define():
    x_in_s = _x_in_s()
    op_def = nodes.OperatorDefinition(
        name='A',
        definiens=x_in_s)
    definitions = [op_def]
    define = nodes.Define(
        definitions=definitions)
    text = pformat(define)
    indent = 8 * '\x20'
    text_ = (
        'DEFINE\n'
        rf'{indent}A == x \in S')
    assert text == text_, text


def test_except():
    # dot
    function = _opname('f')
    item = ['x']
    expression = _x_eq_one()
    change = nodes.FunctionChange(
        item=item,
        expression=expression)
    changes = [change]
    except_ = nodes.Except(
        function=function,
        changes=changes)
    text = pformat(except_)
    text_ = '[f EXCEPT !.x = x = 1]'
    assert text == text_, text
    # brackets
    string_literal = nodes.StringLiteral(
        value='"x"')
    item = [
        [string_literal],
        ]
    change = nodes.FunctionChange(
        item=item,
        expression=expression)
    changes = [change]
    except_ = nodes.Except(
        function=function,
        changes=changes)
    text = pformat(except_)
    text_ = '[f EXCEPT !["x"] = x = 1]'
    assert text == text_, text


def test_fairness():
    operator = 'SF_'
    subscript = _opname('x')
    action = _opname('A')
    fairness = nodes.Fairness(
        operator=operator,
        subscript=subscript,
        action=action)
    text = pformat(fairness)
    text_ = 'SF_x(A)'
    assert text == text_, text


def test_field():
    expression = _opname('f')
    name = 'field_name'
    field = nodes.Field(
        expression=expression,
        name=name)
    text = pformat(field)
    text_ = 'f.field_name'
    assert text == text_, text


def test_float_numeral():
    numeral = '1.2'
    float_numeral = nodes.FloatNumeral(
        value=numeral)
    text = pformat(float_numeral)
    assert text == numeral, text


def test_function_pformat():
    x = _opname('x')
    y = _opname('y')
    s = _opname('S')
    arguments = [y, s]
    y_in_s = nodes.OperatorApplication(
        operator=r'\in',
        arguments=arguments)
    decl_exprs = [x, y_in_s]
    func_value = nodes.OperatorApplication(
        operator='+',
        arguments=[x, y])
    func = nodes.Function(
        declaration=decl_exprs,
        value=func_value)
    text = pformat(func)
    text_ = r'[x, y \in S |-> x + y]'
    assert text == text_, text


def test_function_application():
    func = _opname('f')
    x = _opname('x')
    y = _opname('y')
    arguments = [x, y]
    tree = nodes.FunctionApplication(
        function=func,
        arguments=arguments)
    text = pformat(tree)
    text_ = 'f[x, y]'
    assert text == text_, text


def test_function_change():
    exprs = [_opname('x')]
    name = 'y'
    item = [exprs, name]
    expression = _opname('z')
    change = nodes.FunctionChange(
        item=item,
        expression=expression)
    text = pformat(change)
    text_ = '![x].y = z'
    assert text == text_, text


def test_have():
    expr = _x_in_s()
    have = nodes.Have(
        expression=expr)
    text = pformat(have)
    text_ = r'HAVE x \in S'
    assert text == text_, text


def test_hide():
    theorem_name = _opname('theorem')
    lemma_name = _opname('lemma')
    op_name = _opname('Op')
    facts = [
        theorem_name,
        lemma_name]
    names = [op_name]
    hide = nodes.Hide(
        facts=facts,
        names=names)
    text = pformat(hide)
    text_ = 'HIDE theorem, lemma DEF Op'
    assert text == text_, text


def test_if_pformat():
    predicate = _opname('x')
    then = _opname('y')
    else_ = _opname('z')
    if_ = nodes.If(
        predicate=predicate,
        then=then,
        else_=else_)
    text = pformat(if_)
    text_ = 'IF x THEN y ELSE z'
    assert text == text_, text


def test_ifs():
    x = _opname('x')
    y = _opname('y')
    int_1 = nodes.IntegralNumeral(value='1')
    int_2 = nodes.IntegralNumeral(value='2')
    int_3 = nodes.IntegralNumeral(value='3')
    x_eq_1 = nodes.OperatorApplication(
        operator='=',
        arguments=[x, int_1])
    x_eq_2 = nodes.OperatorApplication(
        operator='=',
        arguments=[x, int_2])
    x_eq_3 = nodes.OperatorApplication(
        operator='=',
        arguments=[x, int_3])
    case_1 = nodes.IfItem(
        predicate=x,
        expression=x_eq_1)
    case_2 = nodes.IfItem(
        predicate=y,
        expression=x_eq_2)
    cases = [case_1, case_2]
    ifs_tree = nodes.Ifs(
        cases=cases,
        other=x_eq_3)
    text = pformat(ifs_tree)
    text_ = (
        'IF x THEN x = 1\n'
        '    ELIF y THEN x = 2\n'
        '    ELSE x = 3')
    assert text == text_, text


def test_if_item():
    x = _opname('x')
    x_in_s = _x_in_s()
    case = nodes.IfItem(
        predicate=x,
        expression=x_in_s)
    text = pformat(case)
    text_ = r'ELIF x THEN x \in S'
    assert text == text_, text


def test_instance():
    name = 'Module'
    instance = nodes.Instance(
        name=name)
    text = pformat(instance)
    text_ = 'INSTANCE Module'
    assert text == text_, text
    # WITH substitutions
    y = _opname('y')
    u = _opname('u')
    v = _opname('v')
    z_expr = nodes.OperatorApplication(
        operator='z',
        arguments=[u, v])
    substitute_x = nodes.WithInstantiation(
        name='x',
        expression=y)
    substitute_y = nodes.WithInstantiation(
        name='y',
        expression=z_expr)
    substitutions = [
        substitute_x,
        substitute_y]
    instance = nodes.Instance(
        name=name,
        with_substitution=substitutions)
    text = pformat(instance)
    text_ = (
        'INSTANCE Module WITH '
        'x <- y, y <- z(u, v)')
    assert text == text_, text
    # local
    instance = nodes.Instance(
        name=name,
        with_substitution=substitutions,
        local=True)
    text = pformat(instance)
    text_ = (
        'LOCAL INSTANCE Module WITH '
        'x <- y, y <- z(u, v)')
    assert text == text_, text


def test_integral_numeral():
    int_1 = nodes.IntegralNumeral(
        value='1')
    text = pformat(int_1)
    assert text == '1', text


def test_key_bound():
    s = _opname('S')
    tree = nodes.KeyBound(
        name='x',
        bound=s)
    text = pformat(tree)
    text_ = 'x: S'
    assert text == text_, text


def test_lambda():
    x = _opname('x')
    y = _opname('y')
    parameters = [x, y]
    expression = _x_in_s()
    tree = nodes.Lambda(
        parameters=parameters,
        expression=expression)
    text = pformat(tree)
    text_ = r'LAMBDA x, y:  x \in S'
    assert text == text_, text


def test_let():
    int_1 = nodes.IntegralNumeral(
        value='1')
    b_def = nodes.OperatorDefinition(
        name='b',
        definiens=int_1)
    definitions = [b_def]
    int_2 = nodes.IntegralNumeral(
        value='2')
    b = _opname('b')
    arguments = [int_2, b]
    expression = nodes.OperatorApplication(
        operator='*',
        arguments=arguments)
    let = nodes.Let(
        definitions=definitions,
        expression=expression)
    text = pformat(let)
    text_ = (
        '\nLET\n'
        '    b == 1\n'
        'IN\n'
        '    2 * b')
    assert text == text_, text


def test_module():
    int_1 = nodes.IntegralNumeral(
        value='1')
    a_def = nodes.OperatorDefinition(
        name='A',
        definiens=int_1)
    extendees = ['Integers']
    units = [a_def]
    module = nodes.Module(
        name='Module',
        extendees=extendees,
        units=units)
    text = pformat(module)
    text_ = (
        '--------------------------------'
        ' MODULE Module '
        '---------------------------------\n'
        'EXTENDS\n'
        '    Integers\n'
        'A == 1\n'
        '========================================'
        '========================================')
    assert text == text_, text


def test_module_name():
    module_name = 'Module'
    tree = nodes.ModuleName(
        name=module_name)
    text = pformat(tree)
    text_ = f'MODULE Module'
    assert text == text_, text


def test_obvious():
    tree = nodes.Obvious()
    text = pformat(tree)
    text_ = '    OBVIOUS'
    assert text == text_, text


def test_omitted():
    tree = nodes.Omitted()
    text = pformat(tree)
    text_ = '    OMITTED'
    assert text == text_, text


def test_operator_application():
    # prefix minus
    x = _opname('x')
    arguments = [x]
    tree = nodes.OperatorApplication(
        operator='-',
        arguments=arguments)
    text = pformat(tree)
    text_ = '- x'
    assert text == text_, text
    # infix minus
    y = _opname('y')
    arguments = [x, y]
    tree = nodes.OperatorApplication(
        operator='-',
        arguments=arguments)
    text = pformat(tree)
    text_ = 'x - y'
    assert text == text_, text
    # postfix `^*`
    arguments = [y]
    tree = nodes.OperatorApplication(
        operator='^*',
        arguments=arguments)
    text = pformat(tree)
    text_ = 'y^*'
    assert text == text_, text
    # infix conjunction
    arguments = [x, y]
    tree = nodes.OperatorApplication(
        operator='/\\',
        arguments=arguments)
    text = pformat(tree)
    text_ = r'x /\ y'
    assert text == text_, text
    # between-operator `\X`
    z = _opname('z')
    arguments = [x, y, z]
    tree = nodes.OperatorApplication(
        operator=r'\X',
        arguments=arguments)
    text = pformat(tree)
    text_ = r'x \X y \X z'
    assert text == text_, text
    # nonfix operator
    tree = nodes.OperatorApplication(
        operator='Op',
        arguments=arguments)
    text = pformat(tree)
    text_ = r'Op(x, y, z)'
    assert text == text_, text


def test_operator_declaration():
    # nullary
    x = _opname('x')
    tree = nodes.OperatorDeclaration(
        name=x,
        level='CONSTANT')
    text = pformat(tree)
    text_ = 'CONSTANT x'
    assert text == text_, text
    # binary
    arguments = ['_', '_']
    y_binary = nodes.OperatorApplication(
        operator='y',
        arguments=arguments)
    tree = nodes.OperatorDeclaration(
        name=y_binary,
        level='ACTION',
        new_keyword=True)
    text = pformat(tree)
    text_ = 'NEW ACTION y(_, _)'
    assert text == text_, text
    # with operator as parameter
    arguments = [y_binary]
    op_arity = nodes.OperatorApplication(
        operator='Op',
        arguments=arguments)
    tree = nodes.OperatorDeclaration(
        name=op_arity,
        level='CONSTANT')
    text = pformat(tree)
    text_ = 'CONSTANT Op(y(_, _))'
    assert text == text_, text


def test_operator_definition():
    name = 'Op'
    x = _opname('x')
    y = _opname('y')
    arguments = [x, y]
    definiens = nodes.OperatorApplication(
        operator='++',
        arguments=arguments)
    x = _opname('x')  # to ensure tree
    y = _opname('y')
    arity = [x, y]
    tree = nodes.OperatorDefinition(
        name=name,
        arity=arity,
        definiens=definiens)
    text = pformat(tree)
    text_ = 'Op(x, y) == x ++ y'
    assert text == text_, text
    # `LOCAL` definition
    x = _opname('x')
    y = _opname('y')
    arguments = [x, y]
    expr = nodes.OperatorApplication(
        operator='<=>',
        arguments=arguments)
    definiens = nodes.Parentheses(
        expression=expr)
    x = _opname('x')  # to ensure tree
    y = _opname('y')
    arity = [x, y]
    tree = nodes.OperatorDefinition(
        name=name,
        arity=arity,
        definiens=definiens,
        local=True)
    text = pformat(tree)
    text_ = 'LOCAL Op(x, y) == (x <=> y)'
    assert text == text_, text


def test_parameter_declaration():
    tree = nodes.ParameterDeclaration(
        name='Op',
        arity=2)
    text = pformat(tree)
    text_ = 'Op(_, _)'
    assert text == text_, text


def test_parentheses():
    x = _opname('x')
    tree = nodes.Parentheses(
        expression=x)
    text = pformat(tree)
    text_ = '(x)'
    assert text == text_, text


def test_pick():
    declarations = [_x_in_s()]
    predicate = _x_eq_one()
    tree = nodes.Pick(
        declarations=declarations,
        predicate=predicate)
    text = pformat(tree)
    text_ = r'PICK x \in S:  x = 1'
    assert text == text_, text


def test_proof():
    # step 1
    name = '<1>1.'
    main = _x_eq_one()
    facts = [_x_in_s()]
    proof = nodes.By(
        facts=facts)
    step_1 = nodes.ProofStep(
        name=name,
        main=main,
        proof=proof)
    # step 2
    name = '<1>2.'
    x = _opname('x')
    int_0 = nodes.IntegralNumeral(
        value='0')
    arguments = [x, int_0]
    main = nodes.OperatorApplication(
        operator='>',
        arguments=arguments)
    facts = [_opname('<1>1')]
    proof = nodes.By(
        facts=facts)
    step_2 = nodes.ProofStep(
        name=name,
        main=main,
        proof=proof,
        proof_keyword=True)
    # proof
    steps = [step_1, step_2]
    tree = nodes.Proof(
        steps=steps,
        proof_keyword=True)
    text = pformat(tree)
    text_ = (
        '    PROOF\n'
        '    <1>1. x = 1\n'
        '        BY x \\in S\n'
        '    <1>2. x > 0\n'
        '        PROOF\n'
        '        BY <1>1')
    assert text == text_


def test_proof_case():
    expression = _x_in_s()
    tree = nodes.ProofCase(
        expression=expression)
    text = pformat(tree)
    text_ = r'CASE x \in S'
    assert text == text_, text


def test_proof_step():
    # with `PROOF` keyword
    name = '<1>1.'
    main = _x_eq_one()
    facts = [_x_in_s()]
    proof = nodes.By(
        facts=facts)
    tree = nodes.ProofStep(
        name=name,
        main=main,
        proof=proof,
        proof_keyword=True)
    text = pformat(tree)
    text_ = (
        '<1>1. x = 1\n'
        '    PROOF\n'
        r'    BY x \in S')
    assert text == text_, text
    # without `PROOF` keyword
    tree = nodes.ProofStep(
        name=name,
        main=main,
        proof=proof)
    text = pformat(tree)
    text_ = (
        '<1>1. x = 1\n'
        r'    BY x \in S')
    # QED proof-step
    name = '<1>5.'
    tree = nodes.ProofStep(
        name=name,
        main='QED',
        proof=proof)
    text = pformat(tree)
    text_ = (
        '<1>5. QED\n'
        r'    BY x \in S')


def test_quantification():
    quantifier = r'\A'
    declarations = [_x_in_s()]
    x = _opname('x')
    int_2 = nodes.FloatNumeral(
        value='2.0')
    arguments = [x, int_2]
    predicate = nodes.OperatorApplication(
        operator='>',
        arguments=arguments)
    tree = nodes.Quantification(
        quantifier=quantifier,
        declarations=declarations,
        predicate=predicate)
    text = pformat(tree)
    text_ = r'\A x \in S:  x > 2.0'
    assert text == text_, text


def test_quantifier_witness():
    x_in_s = _x_in_s()
    y = _opname('y')
    r = _opname('R')
    arguments = [y, r]
    y_in_r = nodes.OperatorApplication(
        operator=r'\in',
        arguments=arguments)
    expressions = [x_in_s, y_in_r]
    tree = nodes.QuantifierWitness(
        expressions=expressions)
    text = pformat(tree)
    text_ = r'WITNESS x \in S, y \in R'
    assert text == text_, text


def test_record():
    int_1 = nodes.IntegralNumeral(
        value=1)
    abc = nodes.StringLiteral(
        value='"abc"')
    key_values = [
        ('x', int_1),
        ('y', abc)]
    tree = nodes.Record(
        key_values=key_values)
    text = pformat(tree)
    text_ = '[x |-> 1, y |-> "abc"]'
    assert text == text_, text


def test_recursive_operators():
    names = [
        _opname('F'),
        _opname('G')]
    tree = nodes.RecursiveOperators(
        names=names)
    text = pformat(tree)
    text_ = 'RECURSIVE F, G'
    assert text == text_, text


def test_sequent():
    assume = [_opname('theorem_name')]
    prove = _x_in_s()
    tree = nodes.Sequent(
        assume=assume,
        prove=prove)
    text = pformat(tree)
    text_ = (
        'ASSUME\n'
        '    theorem_name\n'
        'PROVE\n'
        r'    x \in S')
    assert text == text_, text


def test_set_comprehension():
    arguments = [
        _opname('x'),
        _opname('y')]
    item = nodes.OperatorApplication(
        operator='^',
        arguments=arguments)
    declarations = [
        _x_in_s()]
    tree = nodes.SetComprehension(
        item=item,
        declarations=declarations)
    text = pformat(tree)
    text_ = r'{x ^ y:  x \in S}'
    assert text == text_, text


def test_set_enumeration():
    string = nodes.StringLiteral(
        value='"string literal"')
    integer = nodes.IntegralNumeral(
        value='100')
    float = nodes.FloatNumeral(
        value='1.00')
    items = [string, integer, float]
    tree = nodes.SetEnumeration(
        items=items)
    text = pformat(tree)
    text_ = '{"string literal", 100, 1.00}'
    assert text == text_, text


def test_set_of_booleans():
    booleans = nodes.SetOfBooleans()
    text = pformat(booleans)
    text_ = 'BOOLEAN'
    assert text == text_, text


def test_set_of_functions():
    domain = _opname('dom')
    codomain = _opname('codom')
    tree = nodes.SetOfFunctions(
        domain=domain,
        codomain=codomain)
    text = pformat(tree)
    text_ = '[dom -> codom]'
    assert text == text_, text


def test_set_of_records():
    x_s = nodes.KeyBound(
        name='x',
        bound=_opname('S'))
    y_r = nodes.KeyBound(
        name='y',
        bound=_opname('R'))
    key_bounds = [x_s, y_r]
    tree = nodes.SetOfRecords(
        key_bounds=key_bounds)
    text = pformat(tree)
    text_ = '[x: S, y: R]'
    assert text == text_, text


def test_set_of_strings():
    strings = nodes.SetOfStrings()
    text = pformat(strings)
    text_ = 'STRING'
    assert text == text_, text


def test_set_slice():
    declaration = _x_in_s()
    x = nodes.OperatorApplication(
        operator='x')
    one = nodes.IntegralNumeral(
        value='1')
    arguments = [x, one]
    predicate = nodes.OperatorApplication(
        operator='=',
        arguments=arguments)
    slice = nodes.SetSlice(
        declaration=declaration,
        predicate=predicate)
    text = pformat(slice)
    text_ = r'{x \in S:  x = 1}'
    assert text == text_, text


def test_string_literal():
    string_literal = '"string literal"'
    string = nodes.StringLiteral(
        value=string_literal)
    text = pformat(string)
    assert text == string_literal, text


def test_subexpression_reference():
    # without arguments
    items = ['<1>1', '2', ':']
    tree = nodes.SubexpressionReference(
        items=items)
    text = pformat(tree)
    text_ = '<1>1!2!:'
    assert text == text_, text
    # with arguments
    arguments = [
        _opname('x'),
        _opname('y')]
    op_app = nodes.OperatorApplication(
        operator=None,
        arguments=arguments)
    items = ['<1>2', '3', op_app]
    tree = nodes.SubexpressionReference(
        items=items)
    text = pformat(tree)
    text_ = '<1>2!3!(x, y)'
    assert text == text_, text


def test_subscripted_action():
    # `[...]_...`
    operator = '['
    action = _x_eq_one()
    subscript = _opname('x')
    tree = nodes.SubscriptedAction(
        operator=operator,
        action=action,
        subscript=subscript)
    text = pformat(tree)
    text_ = '[x = 1]_x'
    assert text == text_, text
    # `<<...>>_...`
    operator = '<<'
    tree = nodes.SubscriptedAction(
        operator=operator,
        action=action,
        subscript=subscript)
    text = pformat(tree)
    text_ = r'<<x = 1>>_x'
    assert text == text_, text


def test_suffices():
    predicate = _x_in_s()
    suffices = nodes.Suffices(
        predicate=predicate)
    text = pformat(suffices)
    text_ = (
        'SUFFICES\n'
        r'        x \in S')
    assert text == text_, text


def test_take():
    declarations = [
        _x_in_s(),
        _x_in_s()]
    tree = nodes.Take(
        declarations=declarations)
    text = pformat(tree)
    text_ = r'TAKE x \in S, x \in S'
    assert text == text_, text


def test_temporal_quantification():
    quantifier = r'\EE'
    declarations = [
        _opname('x'),
        _opname('y')]
    expr = _x_eq_one()
    expr = nodes.Parentheses(
        expression=expr)
    arguments = [expr]
    predicate = nodes.OperatorApplication(
        operator='[]',
        arguments=arguments)
    tree = nodes.TemporalQuantification(
        quantifier=quantifier,
        declarations=declarations,
        predicate=predicate)
    text = pformat(tree)
    text_ = r'\EE x, y:  [] (x = 1)'
    assert text == text_, text


def test_theorem():
    # named theorem
    name = 'theorem_name'
    goal = _x_in_s()
    tree = nodes.Theorem(
        name=name,
        goal=goal,
        theorem_keyword='LEMMA')
    text = pformat(tree)
    text_ = (
        'LEMMA theorem_name ==\n'
        r'    x \in S')
    assert text == text_, text
    # theorem with proof
    facts = [_x_eq_one()]
    proof = nodes.By(
        facts=facts)
    tree = nodes.Theorem(
        goal=goal,
        proof=proof,
        theorem_keyword='THEOREM')
    text = pformat(tree)
    text_ = (
        'THEOREM\n'
        '    x \\in S\n'
        'BY x = 1')
    assert text == text_, text
    # keyword error
    tree = nodes.Theorem(
        name=name,
        goal=goal,
        theorem_keyword='LEMMAS')
    with pytest.raises(ValueError):
        pformat(tree)


def test_tuple():
    abc = nodes.StringLiteral(
        value='"abc"')
    one = nodes.IntegralNumeral(
        value='1')
    one_two = nodes.FloatNumeral(
        value='1.2')
    items = [abc, one, one_two]
    tuple_ = nodes.Tuple(
        items=items)
    text = pformat(tuple_)
    text_ = '<<"abc", 1, 1.2>>'
    assert text == text_, text


def test_use():
    z = _opname('z')
    int_2 = nodes.IntegralNumeral(
        value='2')
    int_3 = nodes.IntegralNumeral(
        value='3')
    z_plus_2 = nodes.OperatorApplication(
        operator='+',
        arguments=[z, int_2])
    z_expr = nodes.OperatorApplication(
        operator='=',
        arguments=[z_plus_2, int_3])
    r = _opname('r')
    r_prime = nodes.OperatorApplication(
        operator="'",
        arguments=[r])
    r = _opname('r')  # tree property
    r_expr = nodes.OperatorApplication(
        operator='#',
        arguments=[r_prime, r])
    facts = [z_expr, r_expr]
    names = [_opname('Op')]
    tree = nodes.Use(
        facts=facts,
        names=names)
    text = pformat(tree)
    text_ = "USE z + 2 = 3, r' # r DEF Op"
    assert text == text_, text


def test_variables():
    var_names = ['x', 'y', 'z']
    variables = nodes.Variables(
        names=var_names)
    text = pformat(variables)
    text_ = 'VARIABLE\n    x,\n    y,\n    z'
    assert text == text_, text


def test_vertical_list():
    expression = _x_eq_one()
    item_1 = nodes.VerticalItem(
        operator='\\/',
        expression=expression)
    expression = _x_in_s()
    item_2 = nodes.VerticalItem(
        operator='\\/',
        expression=expression)
    arguments = [item_1, item_2]
    tree = nodes.VerticalList(
        operator='\\/',
        arguments=arguments)
    text = pformat(tree)
    text_ = (
        '\n\\/ x = 1\n'
        '\\/ x \\in S\n')
    assert text == text_, text


def test_vertical_item():
    expression = _x_eq_one()
    tree = nodes.VerticalItem(
        operator='\\/',
        expression=expression)
    text = pformat(tree)
    text_ = 'x = 1'
    assert text == text_, text


def test_with_instantiation():
    name = 'x'
    y = _opname('y')
    int_2 = nodes.IntegralNumeral(
        value='2')
    arguments = [y, int_2]
    expr = nodes.OperatorApplication(
        operator='-',
        arguments=arguments)
    with_inst = nodes.WithInstantiation(
        name=name,
        expression=expr)
    text = pformat(with_inst)
    text_ = 'x <- y - 2'
    assert text == text_, text


def _x_eq_one(
        ) -> tuple:
    """Return tree for `x = 1`."""
    x = nodes.OperatorApplication(
        operator='x')
    one = nodes.IntegralNumeral(
        value='1')
    arguments = [x, one]
    return nodes.OperatorApplication(
        operator='=',
        arguments=arguments)


def _x_in_s(
        ) -> tuple:
    r"""Return tree for `x \in S`."""
    x = nodes.OperatorApplication(
        operator='x')
    bound = nodes.OperatorApplication(
        operator='S')
    arguments = [x, bound]
    return nodes.OperatorApplication(
        operator='\\in',
        arguments=arguments)


def _boolean_expr(
        ) -> tuple:
    """Return tree for `TRUE`."""
    return nodes.BooleanLiteral(
        value='TRUE')


def _wrap_as_module(
        expr_tree:
            tuple
        ) -> tuple:
    """Return tree for module.

    The module has one definition,
    whose definiens is `expr_tree`.
    """
    op_name = 'A'
    def_unit = nodes.OperatorDefinition(
        name=op_name,
        definiens=expr_tree)
    units = [expr_tree]
    module_name = nodes.ModuleName(
        name='module_name')
    return nodes.Module(
        name=module_name,
        extendees=None,
        units=units)


def _opname(
        name:
            str
        ) -> tuple:
    """Return tree for operator name."""
    return nodes.OperatorApplication(
        operator=name)


if __name__ == '__main__':
    test_with_instantiation()
